#!/usr/bin/env python3

from typing import List

from .transaction import (
    Transaction, NormalTransaction, AdvancedTransaction
)

__all__: List[str] = [
    "Transaction",
    "NormalTransaction",
    "AdvancedTransaction"
]



#!/usr/bin/env python3

from binascii import (hexlify, unhexlify)
from mnemonic import Mnemonic
from typing import (Optional, List)

import hmac
import unicodedata
import hashlib

from ..config import config
from ..signature import (sign, verify)
from ..utils import (
    get_mnemonic_language, is_mnemonic, is_network, get_entropy_strength, get_mnemonic_strength
)
from ..exceptions import (
    NetworkError, DerivationError, ClientError
)
from ..rpc import (
    account_create, get_balance, get_utxos
)
from .tools import (
    get_xpublic_key, get_address, get_program, get_child_xpublic_key, get_child_xprivate_key,
    get_private_key, get_public_key, get_expand_xprivate_key
)
from .utils import (
    prune_root_scalar, get_bytes, bad_seed_checker
)

# Bytom config
config: dict = config()


class Wallet:
    """
    Bytom wallet class.

    :param network: Bytom network, defaults to solonet.
    :type network: str
    :returns: Wallet -- Bytom wallet instance.

    .. note::
        Bytom has only three networks, ``mainnet``, ``solonet`` and ``testnet``.
    """

    def __init__(self, network: str = config["network"]):

        if not is_network(network=network):
            raise NetworkError(f"Invalid '{network}' network",
                               "choose only 'mainnet', 'solonet' or 'testnet' networks.")
        self.network: str = network

        self._strength: Optional[int] = None
        self._entropy: Optional[str] = None
        self._mnemonic: Optional[str] = None
        self._language: Optional[str] = None
        self._passphrase: Optional[str] = None

        self._seed: Optional[str] = None
        self._guid: Optional[str] = None
        self._xprivate_key: Optional[str] = None
        self._private_key: Optional[str] = None

        self._indexes: List[str] = []
        self._path: Optional[str] = None

    def from_entropy(self, entropy: str, passphrase: Optional[str] = None,
                     language: str = "english") -> "Wallet":
        """
        Initialize Bytom wallet from entropy.

        :param entropy: Entropy hex string.
        :type entropy: str.
        :param passphrase: Secret password/passphrase, default to None.
        :type passphrase: str.
        :param language: Mnemonic language, default to english.
        :type language: str.
        :returns: Wallet -- Bytom wallet class instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy("8d7454bb99e8e68de6adfc5519cbee64")
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        if language not in ["english", "french", "italian", "japanese",
                            "chinese_simplified", "chinese_traditional", "korean", "spanish"]:
            raise ValueError("Invalid language, choose only the following options 'english', 'french', 'italian', "
                             "'spanish', 'chinese_simplified', 'chinese_traditional', 'japanese or 'korean' languages.")

        self._entropy = unhexlify(entropy.encode())
        self._passphrase, self._language = (
            (passphrase if passphrase is not None else str()), language
        )
        self._strength = get_entropy_strength(entropy=hexlify(self._entropy).decode())
        mnemonic = Mnemonic(language=self._language).to_mnemonic(data=self._entropy)
        self._mnemonic = unicodedata.normalize("NFKC", mnemonic)
        self._seed = Mnemonic.to_seed(mnemonic=self._mnemonic, passphrase=self._passphrase)
        self.from_seed(seed=hexlify(self._seed).decode())
        return self

    def from_mnemonic(self, mnemonic: str, passphrase: Optional[str] = None,
                      language: Optional[str] = None) -> "Wallet":
        """
        Initialize Bytom wallet from mnemonic.

        :param mnemonic: Mnemonic words.
        :type mnemonic: str.
        :param passphrase: Secret password/passphrase, default to None.
        :type passphrase: str.
        :param language: Mnemonic language, default to None.
        :type language: str.
        :returns: Wallet -- Bytom wallet class instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("舒 使 系 款 株 擾 麼 鄉 狗 振 誤 謀")
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        if language and language not in ["english", "french", "italian", "japanese",
                                         "chinese_simplified", "chinese_traditional", "korean", "spanish"]:
            raise ValueError("Invalid language, choose only the following options 'english', 'french', 'italian', "
                             "'spanish', 'chinese_simplified', 'chinese_traditional', 'japanese or 'korean' languages.")
        if not is_mnemonic(mnemonic=mnemonic, language=language):
            raise ValueError("Invalid mnemonic words.")

        self._mnemonic = unicodedata.normalize("NFKC", mnemonic)
        self._language, self._passphrase = (
            get_mnemonic_language(mnemonic=self._mnemonic),
            (passphrase if passphrase is not None else str())
        )
        self._entropy = Mnemonic(language=self._language).to_entropy(self._mnemonic)
        self._strength = get_mnemonic_strength(mnemonic=self._mnemonic)
        self._seed = Mnemonic.to_seed(
            mnemonic=self._mnemonic, passphrase=self._passphrase)
        self.from_seed(seed=hexlify(self._seed).decode())
        return self

    def from_seed(self, seed: str) -> "Wallet":
        """
        Initialize Bytom wallet from seed.

        :param seed: Mnemonic seed hex string.
        :type seed: str.
        :returns: Wallet -- Bytom wallet class instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_seed("baff3e1fe60e1f2a2d840d304acc98d1818140c79354a353b400fb019bfb256bc392d7aa9047adff1f14bce0342e14605c6743a6c08e02150588375eb2eb7d49")
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        self._seed = unhexlify(seed.encode())
        i = hmac.HMAC(b"Root", get_bytes(self._seed),
                      digestmod=hashlib.sha512).hexdigest()
        il, ir = i[:64], i[64:]
        bad_seed_checker(il, True)

        # get root xprivate_key key
        self._xprivate_key = prune_root_scalar(il).hex() + ir
        self._private_key = get_private_key(str(self._xprivate_key), self._indexes)
        return self

    def from_xprivate_key(self, xprivate_key: str) -> "Wallet":
        """
        Initialize Bytom wallet from xprivate key.

        :param xprivate_key: Bytom xprivate key.
        :type xprivate_key: str.
        :returns: Wallet -- Bytom wallet class instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_xprivate_key("205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee51ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b")
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        self._xprivate_key = xprivate_key
        return self

    def from_private_key(self, private_key: str) -> "Wallet":
        """
        Initialize Bytom wallet from private key.

        :param private_key: Bytom private key.
        :type private_key: str.
        :returns: Wallet -- Bytom wallet class instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_private_key("e0d42c3a1d9e1c54c09d5da9fd582afb1d053d3c033c3a07fedf2a709ce3f4477b4b52132f610150767edac6e1c2934d34780a9340a56a9dea58e070e44b70f1")
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        self._private_key = private_key
        return self

    def derivation(self, index: Optional[int] = None) -> "Wallet":
        if index is not None:
            index = int(index).to_bytes(4, byteorder="little").hex()
            self._indexes.append(index)
        if not self._xprivate_key:
            raise DerivationError("You can't drive", "XPrivate Key is also None.")
        self._private_key = get_private_key(str(self._xprivate_key), self._indexes)
        return self

    def from_indexes(self, indexes: List[str]) -> "Wallet":
        """
        Drive Bytom wallet from indexes.

        :param indexes: Bytom derivation indexes.
        :type indexes: list.
        :returns: Wallet -- Bytom wallet class instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_xprivate_key("205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee51ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        self._indexes = indexes
        self.derivation()
        return self

    def from_index(self, index: int, harden: bool = False) -> "Wallet":
        """
        Drive Bytom wallet from index.

        :param index: Bytom derivation index.
        :type index: int.
        :param harden: BIP 32 key harden, default to False.
        :type harden: bool.
        :returns: Wallet -- Bytom wallet class instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy(entropy="8d7454bb99e8e68de6adfc5519cbee64", language="japaneses")
        >>> wallet.from_index(44)
        >>> wallet.from_index(153)
        >>> wallet.from_index(1)  # Account
        >>> wallet.from_index(0)  # Change False(0)/True(1)
        >>> wallet.from_index(1)  # Address
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        if harden:
            self.derivation(index + 0x80000000)
        else:
            self.derivation(index)
        return self

    def from_path(self, path: str) -> "Wallet":
        """
        Drive Bytom wallet from path.

        :param path: Bytom derivation path.
        :type path: str.
        :returns:  Wallet -- Bytom wallet instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic(mnemonic="舒 使 系 款 株 擾 麼 鄉 狗 振 誤 謀", passphrase="Hello Meheret!")
        >>> wallet.from_path("m/44/153/1/0/1")
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        if str(path)[0:2] != "m/":
            raise DerivationError("Bad path", "insert like this type of path \"m/0'/0\"! ")

        for index in path.lstrip("m/").split("/"):
            if "'" in index:
                self.derivation(int(index[:-1]) + 0x80000000)
            else:
                self.derivation(int(index))
        return self

    def clean_derivation(self) -> "Wallet":
        """
        Clean derivation indexes/path.

        :returns:  Wallet -- Bytom wallet instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic(mnemonic="舒 使 系 款 株 擾 麼 鄉 狗 振 誤 謀", passphrase="Hello Meheret!")
        >>> wallet.from_path("m/44/153/1/0/1")
        >>> wallet.indexes()
        ["2c000000", "99000000", "01000000", "00000000", "01000000"]
        >>> wallet.path()
        "m/44/153/1/0/1"
        >>> wallet.clean_derivation()
        >>> wallet.indexes()
        []
        >>> wallet.path()
        None
        """

        self._indexes = list()
        self.derivation()
        return self

    def guid(self) -> Optional[str]:
        """
        Get Bytom wallet blockcenter guid.

        :return: str -- Bytom blockcenter guid.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.guid()
        "f0ed6ddd-9d6b-49fd-8866-a52d1083a13b"
        """

        if self._guid is None:
            if not self.xpublic_key():
                raise ClientError("You can't get GUID", "xpublic key is None.")
            self._guid = account_create(
                xpublic_key=self.xpublic_key(), network=self.network)["guid"]
        return self._guid

    def strength(self) -> Optional[int]:
        """
        Get wallet entropy/mnemonic strength.

        :return: int -- Entropy/Mnemonic strength.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy(entropy="8d7454bb99e8e68de6adfc5519cbee64", language="korean")
        >>> wallet.strength()
        128
        """

        return self._strength if self._strength else None

    def entropy(self) -> Optional[str]:
        """
        Get wallet entropy.

        :return: str -- Entropy hex string.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy(entropy="8d7454bb99e8e68de6adfc5519cbee64", language="korean")
        >>> wallet.entropy()
        "8d7454bb99e8e68de6adfc5519cbee64"
        """

        return hexlify(self._entropy).decode() if self._entropy else None

    def mnemonic(self) -> Optional[str]:
        """
        Get wallet mnemonic.

        :return: str -- 12 word mnemonic.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy(entropy="8d7454bb99e8e68de6adfc5519cbee64", language="italian")
        >>> wallet.mnemonic()
        "occasione pizzico coltivato cremoso odorare epilogo patacca salone fonia sfuso vispo selettivo"
        """

        return unicodedata.normalize("NFKC", self._mnemonic) if self._mnemonic else None

    def passphrase(self) -> Optional[str]:
        """
        Get wallet secret password/passphrase.

        :return: str -- Secret password/passphrase.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy(entropy="8d7454bb99e8e68de6adfc5519cbee64", passphrase="Hello Meheret!", language="italian")
        >>> wallet.passphrase()
        "Hello Meheret!"
        """

        return str(self._passphrase) if self._passphrase else None

    def language(self) -> Optional[str]:
        """
        Get wallet mnemonic language.

        :return: str -- Mnemonic language.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy(entropy="8d7454bb99e8e68de6adfc5519cbee64", language="italian")
        >>> wallet.language()
        "italian"
        """

        return self._language if self._language else None

    def seed(self) -> Optional[str]:
        """
        Get wallet mnemonic seed.

        :return: str -- Mnemonic seed.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.seed()
        "baff3e1fe60e1f2a2d840d304acc98d1818140c79354a353b400fb019bfb256bc392d7aa9047adff1f14bce0342e14605c6743a6c08e02150588375eb2eb7d49"
        """

        return hexlify(self._seed).decode() if self._seed else None

    def xprivate_key(self) -> Optional[str]:
        """
        Get Bytom wallet xprivate key.

        :return: str -- Bytom xprivate key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.xprivate_key()
        "205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee51ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b"
        """

        return self._xprivate_key

    def xpublic_key(self) -> Optional[str]:
        """
        Get Bytom wallet xpublic key.

        :return: str -- Bytom xpublic key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.xpublic_key()
        "16476b7fd68ca2acd92cfc38fa353e75d6103f828276f44d587e660a6bd7a5c5ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b"
        """

        return get_xpublic_key(xprivate_key=self.xprivate_key()) if self._xprivate_key else None

    def expand_xprivate_key(self) -> Optional[str]:
        """
        Get Bytom wallet expand xprivate key.

        :return: str -- Bytom expand xprivate key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.expand_xprivate_key()
        "205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee5102416c643cfb46ab1ae5a524c8b4aaa002eb771d0d9cfc7490c0c3a8177e053e"
        """

        return get_expand_xprivate_key(xprivate_key=self.xprivate_key()) if self._xprivate_key else None

    def private_key(self) -> str:
        """
        Get Bytom wallet private key.

        :return: str -- Bytom private key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_path("m/44/153/1/0/1")
        >>> wallet.private_key()
        "e07af52746e7cccd0a7d1fba6651a6f474bada481f34b1c5bab5e2d71e36ee515803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141"
        """

        return self._private_key

    def public_key(self) -> str:
        """
        Get Bytom wallet public key.

        :return: str -- Bytom public key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_path("m/44/153/1/0/1")
        >>> wallet.public_key()
        "91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2"
        """

        return get_public_key(xpublic_key=self.xpublic_key(), indexes=self.indexes()) \
            if self._xprivate_key else get_xpublic_key(xprivate_key=self.private_key())[:64]

    def indexes(self) -> Optional[list]:
        """
        Get Bytom wallet derivation indexes.

        :return: list -- Bytom derivation indexes.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_path("m/44/153/1/0/1")
        >>> wallet.indexes()
        ["2c000000", "99000000", "01000000", "00000000", "01000000"]
        """

        return self._indexes

    def path(self) -> Optional[str]:
        """
        Get Bytom wallet derivation path.

        :return: str -- Bytom derivation path.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        >>> wallet.path()
        "m/44/153/1/0/1"
        """

        path = "m/"
        for i, index in enumerate(self._indexes, 1):
            number = int.from_bytes(bytes.fromhex(index), byteorder="little")
            if i == len(self._indexes):
                path = path + str(number)
            else:
                path = path + str(number) + "/"
        return path if not path == "m/" else None

    def child_xprivate_key(self) -> Optional[str]:
        """
        Get Bytom get child xprivate key.

        :return: str -- Bytom child xprivate key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_path("m/44/153/1/0/1")
        >>> wallet.child_xprivate_key()
        "e07af52746e7cccd0a7d1fba6651a6f474bada481f34b1c5bab5e2d71e36ee515803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141"
        """

        return get_child_xprivate_key(xprivate_key=self.xprivate_key(),
                                      indexes=self.indexes()) if self._xprivate_key else None

    def child_xpublic_key(self) -> Optional[str]:
        """
        Get Bytom get child xpublic key.

        :return: str -- Bytom child xpublic key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_path("m/44/153/1/0/1")
        >>> wallet.child_xpublic_key()
        "91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e25803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141"
        """

        return get_child_xpublic_key(xpublic_key=self.xpublic_key(),
                                     indexes=self.indexes()) if self._xprivate_key else None

    def program(self) -> str:
        """
        Get Bytom wallet control program.

        :return: str -- Bytom wallet control program.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        >>> wallet.program()
        "00142cda4f99ea8112e6fa61cdd26157ed6dc408332a"
        """

        return get_program(public_key=self.public_key())

    def address(self, network: Optional[str] = None, vapor: bool = config["vapor"]) -> str:
        """
        Get Bytom wallet address.

        :param network: Bytom network, defaults to solonet.
        :type network: str
        :param vapor: Bytom sidechain vapor, defaults to False.
        :type vapor: bool
        :return: str -- Bytom wallet address.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        >>> wallet.address(network="mainnet", vapor=False)
        "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7"
        >>> wallet.address(network="mainnet", vapor=True)
        "vp1q9ndylx02syfwd7npehfxz4lddhzqsve2za23ag"
        """

        if network is None:
            network = self.network

        return get_address(program=self.program(), network=network, vapor=vapor)

    def balance(self, asset: str = config["asset"], vapor: bool = config["vapor"]) -> int:
        """
        Get Bytom wallet balance.

        :param asset: Bytom asset, defaults to BTM asset.
        :type asset: str
        :param vapor: Bytom sidechain vapor, defaults to False.
        :type vapor: bool
        :return: int -- Bytom balance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.balance()
        2450000000
        """

        return get_balance(
            address=self.address(network=self.network, vapor=vapor),
            asset=asset, network=self.network, vapor=vapor
        )

    def utxos(self, asset: str = config["asset"], limit: int = 15) -> list:
        """
        Get Bytom wallet unspent transaction output (UTXO's).

        :param asset: Bytom asset id, defaults to BTM asset.
        :type asset: str
        :param limit: Bytom balance, default is 15.
        :type limit: int

        :return: list -- Bytom unspent transaction outputs.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy("50f002376c81c96e430b48f1fe71df57")
        >>> wallet.from_path("m/44/153/1/0/1")
        >>> wallet.utxos()
        [{'hash': 'e152f88d33c6659ad823d15c5c65b2ed946d207c42430022cba9bb9b9d70a7a4', 'asset': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 587639800}, {'hash': '88289fa4c7633574931be7ce4102aeb24def0de20e38e7d69a5ddd6efc116b95', 'asset': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 8160000}, {'hash': 'f71c68f921b434cc2bcd469d26e7927aa6db7500e4cdeef814884f11c10f5de2', 'asset': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 10000}, {'hash': 'e46cfecc1f1a26413172ce81c78affb19408e613915642fa5fb04d3b0a4ffa65', 'asset': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 100}]
        """

        return get_utxos(program=self.program(), asset=asset, limit=limit)

    def sign(self, message: str) -> str:
        """
        Sign message data by private key.

        :param message: Message data.
        :type message: str.
        :return: str -- Signed message data (signature).

        >>> from pybytom.wallet import Wallet
        >>> message = "1246b84985e1ab5f83f4ec2bdf271114666fd3d9e24d12981a3c861b9ed523c6"
        >>> wallet = Wallet(network="testnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        >>> wallet.sign(message=message)
        "f6624fea84fadccbc1bc72dc384f662468e271c4e32d846bc0a1524470549992c8ffcc3ca43891a30de4235392b0868c506ed254f0f77cc1f2b9c1a2385ddb05"
        """

        return sign(private_key=self.private_key(), message=message)

    def verify(self, message: str, signature: str) -> bool:
        """
        Verify signature by public key.

        :param message: Message data.
        :type message: str.
        :param signature: Signed message data.
        :type signature: str.
        :return: bool -- Verified signature (True/False).

        >>> from pybytom.wallet import Wallet
        >>> message = "1246b84985e1ab5f83f4ec2bdf271114666fd3d9e24d12981a3c861b9ed523c6"
        >>> signature = "f6624fea84fadccbc1bc72dc384f662468e271c4e32d846bc0a1524470549992c8ffcc3ca43891a30de4235392b0868c506ed254f0f77cc1f2b9c1a2385ddb05"
        >>> wallet = Wallet(network="testnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        >>> wallet.verify(message=message, signature=signature)
        True
        """

        return verify(public_key=self.public_key(), message=message, signature=signature)

    def dumps(self, guid: bool = False) -> dict:
        """
        Get Bytom all wallet information's

        :param guid: Get GUID, default to False.
        :type guid: bool.
        :return: dict -- Bytom all wallet information's.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet()
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        >>> wallet.dumps()
        {'strength': 128, 'entropy': None, 'mnemonic': 'indicate warm sock mistake code spot acid ribbon sing over taxi toast', 'language': 'english', 'passphrase': None, 'seed': 'baff3e1fe60e1f2a2d840d304acc98d1818140c79354a353b400fb019bfb256bc392d7aa9047adff1f14bce0342e14605c6743a6c08e02150588375eb2eb7d49', 'xprivate_key': '205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee51ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b', 'xpublic_key': '16476b7fd68ca2acd92cfc38fa353e75d6103f828276f44d587e660a6bd7a5c5ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b', 'expand_xprivate_key': '205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee5102416c643cfb46ab1ae5a524c8b4aaa002eb771d0d9cfc7490c0c3a8177e053e', 'guid': None, 'indexes': ['2c000000', '99000000', '01000000', '00000000', '01000000'], 'path': 'm/44/153/1/0/1', 'child_xprivate_key': 'e07af52746e7cccd0a7d1fba6651a6f474bada481f34b1c5bab5e2d71e36ee515803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141', 'child_xpublic_key': '91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e25803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141', 'private_key': 'e07af52746e7cccd0a7d1fba6651a6f474bada481f34b1c5bab5e2d71e36ee515803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141', 'public_key': '91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2', 'program': '00142cda4f99ea8112e6fa61cdd26157ed6dc408332a', 'address': {'mainnet': 'bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7', 'solonet': 'sm1q9ndylx02syfwd7npehfxz4lddhzqsve2gdsdcs', 'testnet': 'tm1q9ndylx02syfwd7npehfxz4lddhzqsve2d2mgc0'}}
        """

        return dict(
            strength=self.strength(),
            entropy=self.entropy(),
            mnemonic=self.mnemonic(),
            language=self.language(),
            passphrase=self.passphrase(),
            seed=self.seed(),
            xprivate_key=self.xprivate_key(),
            xpublic_key=self.xpublic_key(),
            expand_xprivate_key=self.expand_xprivate_key(),
            guid=self.guid() if guid else None,
            indexes=self.indexes(),
            path=self.path(),
            child_xprivate_key=self.child_xprivate_key(),
            child_xpublic_key=self.child_xpublic_key(),
            private_key=self.private_key(),
            public_key=self.public_key(),
            program=self.program(),
            address=dict(
                mainnet=self.address(network="mainnet", vapor=False),
                solonet=self.address(network="solonet", vapor=False),
                testnet=self.address(network="testnet", vapor=False)
            ),
            vapor_address=dict(
                mainnet=self.address(network="mainnet", vapor=True),
                solonet=self.address(network="solonet", vapor=True),
                testnet=self.address(network="testnet", vapor=True)
            )
        )



#!/usr/bin/env python3

from typing import List

OP_FALSE: int = 0x00
OP_TRUE: int = 0x51

OP_0: int = 0x00
OP_1: int = 0x51
OP_2: int = 0x52
OP_3: int = 0x53
OP_4: int = 0x54
OP_5: int = 0x55
OP_6: int = 0x56
OP_7: int = 0x57
OP_8: int = 0x58
OP_9: int = 0x59
OP_10: int = 0x5a
OP_11: int = 0x5b
OP_12: int = 0x5c
OP_13: int = 0x5d
OP_14: int = 0x5e
OP_15: int = 0x5f
OP_16: int = 0x60

OP_DATA_1: int = 0x01
OP_DATA_2: int = 0x02
OP_DATA_3: int = 0x03
OP_DATA_4: int = 0x04
OP_DATA_5: int = 0x05
OP_DATA_6: int = 0x06
OP_DATA_7: int = 0x07
OP_DATA_8: int = 0x08
OP_DATA_9: int = 0x09
OP_DATA_10: int = 0x0a
OP_DATA_11: int = 0x0b
OP_DATA_12: int = 0x0c
OP_DATA_13: int = 0x0d
OP_DATA_14: int = 0x0e
OP_DATA_15: int = 0x0f
OP_DATA_16: int = 0x10
OP_DATA_17: int = 0x11
OP_DATA_18: int = 0x12
OP_DATA_19: int = 0x13
OP_DATA_20: int = 0x14
OP_DATA_21: int = 0x15
OP_DATA_22: int = 0x16
OP_DATA_23: int = 0x17
OP_DATA_24: int = 0x18
OP_DATA_25: int = 0x19
OP_DATA_26: int = 0x1a
OP_DATA_27: int = 0x1b
OP_DATA_28: int = 0x1c
OP_DATA_29: int = 0x1d
OP_DATA_30: int = 0x1e
OP_DATA_31: int = 0x1f
OP_DATA_32: int = 0x20
OP_DATA_33: int = 0x21
OP_DATA_34: int = 0x22
OP_DATA_35: int = 0x23
OP_DATA_36: int = 0x24
OP_DATA_37: int = 0x25
OP_DATA_38: int = 0x26
OP_DATA_39: int = 0x27
OP_DATA_40: int = 0x28
OP_DATA_41: int = 0x29
OP_DATA_42: int = 0x2a
OP_DATA_43: int = 0x2b
OP_DATA_44: int = 0x2c
OP_DATA_45: int = 0x2d
OP_DATA_46: int = 0x2e
OP_DATA_47: int = 0x2f
OP_DATA_48: int = 0x30
OP_DATA_49: int = 0x31
OP_DATA_50: int = 0x32
OP_DATA_51: int = 0x33
OP_DATA_52: int = 0x34
OP_DATA_53: int = 0x35
OP_DATA_54: int = 0x36
OP_DATA_55: int = 0x37
OP_DATA_56: int = 0x38
OP_DATA_57: int = 0x39
OP_DATA_58: int = 0x3a
OP_DATA_59: int = 0x3b
OP_DATA_60: int = 0x3c
OP_DATA_61: int = 0x3d
OP_DATA_62: int = 0x3e
OP_DATA_63: int = 0x3f
OP_DATA_64: int = 0x40
OP_DATA_65: int = 0x41
OP_DATA_66: int = 0x42
OP_DATA_67: int = 0x43
OP_DATA_68: int = 0x44
OP_DATA_69: int = 0x45
OP_DATA_70: int = 0x46
OP_DATA_71: int = 0x47
OP_DATA_72: int = 0x48
OP_DATA_73: int = 0x49
OP_DATA_74: int = 0x4a
OP_DATA_75: int = 0x4b

OP_PUSHDATA1: int = 0x4c
OP_PUSHDATA2: int = 0x4d
OP_PUSHDATA4: int = 0x4e
OP_1NEGATE: int = 0x4f
OP_NOP: int = 0x61

OP_JUMP: int = 0x63
OP_JUMPIF: int = 0x64
OP_VERIFY: int = 0x69
OP_FAIL: int = 0x6a
OP_CHECKPREDICATE: int = 0xc0

OP_TOALTSTACK: int = 0x6b
OP_FROMALTSTACK: int = 0x6c
OP_2DROP: int = 0x6d
OP_2DUP: int = 0x6e
OP_3DUP: int = 0x6f
OP_2OVER: int = 0x70
OP_2ROT: int = 0x71
OP_2SWAP: int = 0x72
OP_IFDUP: int = 0x73
OP_DEPTH: int = 0x74
OP_DROP: int = 0x75
OP_DUP: int = 0x76
OP_NIP: int = 0x77
OP_OVER: int = 0x78
OP_PICK: int = 0x79
OP_ROLL: int = 0x7a
OP_ROT: int = 0x7b
OP_SWAP: int = 0x7c
OP_TUCK: int = 0x7d

OP_CAT: int = 0x7e
OP_SUBSTR: int = 0x7f
OP_LEFT: int = 0x80
OP_RIGHT: int = 0x81
OP_SIZE: int = 0x82
OP_CATPUSHDATA: int = 0x89

OP_INVERT: int = 0x83
OP_AND: int = 0x84
OP_OR: int = 0x85
OP_XOR: int = 0x86
OP_EQUAL: int = 0x87
OP_EQUALVERIFY: int = 0x88

OP_1ADD: int = 0x8b
OP_1SUB: int = 0x8c
OP_2MUL: int = 0x8d
OP_2DIV: int = 0x8e
OP_NEGATE: int = 0x8f
OP_ABS: int = 0x90
OP_NOT: int = 0x91
OP_0NOTEQUAL: int = 0x92
OP_ADD: int = 0x93
OP_SUB: int = 0x94
OP_MUL: int = 0x95
OP_DIV: int = 0x96
OP_MOD: int = 0x97
OP_LSHIFT: int = 0x98
OP_RSHIFT: int = 0x99
OP_BOOLAND: int = 0x9a
OP_BOOLOR: int = 0x9b
OP_NUMEQUAL: int = 0x9c
OP_NUMEQUALVERIFY: int = 0x9d
OP_NUMNOTEQUAL: int = 0x9e
OP_LESSTHAN: int = 0x9f
OP_GREATERTHAN: int = 0xa0
OP_LESSTHANOREQUAL: int = 0xa1
OP_GREATERTHANOREQUAL: int = 0xa2
OP_MIN: int = 0xa3
OP_MAX: int = 0xa4
OP_WITHIN: int = 0xa5

OP_SHA256: int = 0xa8
OP_SHA3: int = 0xaa
OP_HASH160: int = 0xab
OP_CHECKSIG: int = 0xac
OP_CHECKMULTISIG: int = 0xad
OP_TXSIGHASH: int = 0xae

OP_CHECKOUTPUT: int = 0xc1
OP_ASSET: int = 0xc2
OP_AMOUNT: int = 0xc3
OP_PROGRAM: int = 0xc4
OP_INDEX: int = 0xc9
OP_ENTRYID: int = 0xca
OP_OUTPUTID: int = 0xcb
OP_BLOCKHEIGHT: int = 0xcd


# OP_DATAS
def op_datas() -> List[int]:
    return [
        0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d,
        0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a,
        0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
        0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32, 0x33, 0x34,
        0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40, 0x41,
        0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b
    ]



#!/usr/bin/env python3

from .assets import BTM as ASSET


# Bytom mainchain & sidechain configuration
def config():
    return {
        "mainchain": {
            "mainnet": {
                "bytom": "http://localhost:9888",
                "blockmeta": "https://blockmeta.com/api/v3",
                "blockcenter": "https://bcapi.bystack.com/bytom/v3",
                "mov": "https://ex.movapi.com/bytom/v3"
            },
            "solonet": {
                "bytom": "http://localhost:9888",
                "blockmeta": None,
                "blockcenter": None,
                "mov": None
            },
            "testnet": {
                "bytom": "http://localhost:9888",
                "blockmeta": None,
                "blockcenter": None,
                "mov": None
            }
        },
        "sidechain": {
            "mainnet": {
                "bytom": "http://localhost:9888",
                "blockmeta": "https://vapor.blockmeta.com/api/v1",
                "blockcenter": "https://ex.movapi.com/vapor/v3",
                "mov": "https://ex.movapi.com/bytom/v3"
            },
            "solonet": {
                "bytom": "http://localhost:9888",
                "blockmeta": None,
                "blockcenter": None,
                "mov": None
            },
            "testnet": {
                "bytom": "http://localhost:9888",
                "blockmeta": None,
                "blockcenter": None,
                "mov": None
            }
        },
        "harden": 0x80000000,
        "vapor": False,
        "network": "mainnet",
        "timeout": 60,
        "asset": ASSET,
        "symbols": {
            "BTM": 1,
            "mBTM": 1000,
            "NEU": 100_000_000
        },
        "path": "m/44/153/1/0/1",
        "indexes": ["2c000000", "99000000", "01000000", "00000000", "01000000"],
        "fee": 10_000_000,
        "confirmations": 1,
        "forbid_chain_tx": False,
        "headers": {
            "User-Agent": "PyBytom User Agent v0.1.0",
            "Content-Type": "application/json; charset=utf-8",
            "Accept": "application/json"
        }
    }



#!/usr/bin/env python3

from typing import List

# Bytom asset
BTM: str = "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff"
# Bitcoin asset
BTC: str = "bda946b3110fa46fd94346ce3f05f0760f1b9de72e238835bc4d19f9d64f1742"
# Dai Stablecoin asset
DAI: str = "25f2069140fa3ff4d6e0dc1d0fcaa11ace01eb721f115f0f1a5a3782db597fb1"
# Ethereum asset
ETH: str = "78de44ffa1bce37b757c9eae8925b5f199dc4621b412ef0f3f46168865284a93"
# Litecoin asset
LTC: str = "011a24f9da7551d4cd9ae0f194aa1d1691e22a173edf7d81aabd9a97ca386252"
# SUP asset
SUP: str = "47fcd4d7c22d1d38931a6cd7767156babbd5f05bbbb3f7d3900635b56eb1b67e"
# USD Coin asset
USDC: str = "c4644dd6643475d57ed624f63129ab815f282b61f4bb07646d73423a6e1a1563"
# Tether asset
USDT: str = "184e1cc4ee4845023888810a79eed7a42c02c544cf2c61ceac05e176d575bd46"


__all__:  List[str] = [
    BTM, BTC, DAI, ETH, LTC, SUP, USDC, USDT
]



#!/usr/bin/env python3

from typing import List

from .. config import config
from .wallet import Wallet
from .tools import (
    get_xpublic_key, get_expand_xprivate_key, get_child_xprivate_key, indexes_to_path,
    get_child_xpublic_key, get_address, get_program, get_private_key,
    get_public_key, get_bytes, path_to_indexes
)

# Bytom config
config: dict = config()

# Default derivation path
DEFAULT_PATH: str = config["path"]
# Default derivation indexes
DEFAULT_INDEXES: List[str] = config["indexes"]


__all__: List[str] = [
    "Wallet", "DEFAULT_PATH", "DEFAULT_INDEXES",
    "get_xpublic_key", "get_expand_xprivate_key", "get_child_xprivate_key",
    "get_child_xpublic_key", "get_address", "get_program",
    "get_private_key", "get_public_key", "get_bytes",
    "indexes_to_path", "path_to_indexes"
]



#!/usr/bin/env python3

from typing import Optional, List

import hmac
import hashlib

from ..libs.segwit import encode
from ..libs.ed25519 import (
    encodepoint, decodepoint, decodeint, scalarmultbase, edwards_add
)
from .utils import (
    prune_intermediate_scalar, get_bytes, bad_seed_checker
)
from ..exceptions import NetworkError
from ..utils import is_network
from ..config import config

# Bytom config
config: dict = config()


def get_xpublic_key(xprivate_key: str) -> str:
    """
    Get Bytom xpublic key from xprivate key.

    :param xprivate_key: Bytom xprivate key.
    :type xprivate_key: str

    :return: str -- Bytom xpublic key.

    >>> from pybytom.wallet.tools import get_xpublic_key
    >>> get_xpublic_key("205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee51ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b")
    "16476b7fd68ca2acd92cfc38fa353e75d6103f828276f44d587e660a6bd7a5c5ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b"
    """

    xprivate_bytes = get_bytes(xprivate_key)
    scalar = decodeint(xprivate_bytes[:len(xprivate_bytes) // 2])
    buf = encodepoint(scalarmultbase(scalar))
    xpublic_key = buf + xprivate_bytes[len(xprivate_bytes) // 2:]
    return xpublic_key.hex()


def get_expand_xprivate_key(xprivate_key: str) -> str:
    """
    Get Bytom expand xprivate key from xprivate key.

    :param xprivate_key: Bytom xprivate key.
    :type xprivate_key: str

    :return: str -- Bytom expand xprivate key.

    >>> from pybytom.wallet.tools import get_expand_xprivate_key
    >>> get_expand_xprivate_key("205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee51ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b")
    "205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee5102416c643cfb46ab1ae5a524c8b4aaa002eb771d0d9cfc7490c0c3a8177e053e"
    """

    i = hmac.HMAC(b"Expand", get_bytes(xprivate_key),
                  digestmod=hashlib.sha512).hexdigest()
    il, ir = i[:64], i[64:]
    bad_seed_checker(ir, True)

    expand_xprivate_key = xprivate_key[:64] + ir
    return expand_xprivate_key


def indexes_to_path(indexes: List[str]) -> str:
    """
    Change derivation indexes to path.

    :param indexes: Bytom derivation indexes.
    :type indexes: list

    :return: str -- Bytom derivation path.

    >>> from pybytom.wallet.tools import indexes_to_path
    >>> indexes_to_path(["2c000000", "99000000", "01000000", "00000000", "01000000"])
    "m/44/153/1/0/1"
    """

    path = "m/"
    for i, index in enumerate(indexes, 1):
        number = int.from_bytes(bytes.fromhex(index), byteorder="little")
        if i == len(indexes):
            path = path + str(number)
        else:
            path = path + str(number) + "/"
    return path


def path_to_indexes(path: str) -> List[str]:
    """
    Change derivation path to indexes.

    :param path: Bytom derivation path.
    :type path: str

    :return: list -- Bytom derivation indexes.

    >>> from pybytom.wallet.tools import path_to_indexes
    >>> path_to_indexes("m/44/153/1/0/1")
    ["2c000000", "99000000", "01000000", "00000000", "01000000"]
    """

    if str(path)[0:2] != "m/":
        raise ValueError("bad path, insert like this type of path \"m/0'/0\"! ")

    indexes = list()
    for index in path.lstrip("m/").split("/"):
        if "'" in index:
            index = int(int(index[:-1]) + config["harden"]).to_bytes(4, byteorder="little").hex()
            indexes.append(index)
        else:
            index = int(int(index)).to_bytes(4, byteorder="little").hex()
            indexes.append(index)
    return indexes


def get_child_xprivate_key(xprivate_key: str, indexes: Optional[List[str]] = None,
                           path: Optional[str] = None) -> str:
    """
    Get Bytom get child xprivate key.

    :param xprivate_key: Bytom xprivate key.
    :type xprivate_key: str
    :param indexes: Bytom derivation indexes, default to ["2c000000", "99000000", "01000000", "00000000", "01000000"].
    :type indexes: list
    :param path: Bytom derivation path, default to None.
    :type path: str

    :return: str -- Bytom child xprivate key.

    >>> from pybytom.wallet.tools import get_child_xprivate_key
    >>> get_child_xprivate_key("205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee51ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b")
    "00ca8555655d336c4c0d11464a1d401f0cc7c29fdc52bf52f5fc8e0ced32ee51b9c62d145693b366cde5ba74a06962bfa9f6b1e810a3e15eadf791247333547e"
    """

    if indexes is None and path is None:
        indexes = []
    elif path is not None:
        indexes = path_to_indexes(path=path)

    for index in range(len(indexes)):
        index_bytes = get_bytes(indexes[index])
        xpublic_key = get_xpublic_key(xprivate_key=xprivate_key)
        xpublic_bytes = get_bytes(xpublic_key)
        xprivate_bytes = get_bytes(xprivate_key)
        i = bytearray(hmac.HMAC(xpublic_bytes[32:],
                                b"N" + xpublic_bytes[:32] + index_bytes,
                                digestmod=hashlib.sha512).digest())
        il, ir = i[:32], i[32:]
        bad_seed_checker(il)

        i = prune_intermediate_scalar(il)[:32] + ir

        carry = 0
        total = 0
        for _i in range(32):
            total = xprivate_bytes[_i] + i[_i] + carry
            i[_i] = total & 0xff
            carry = total >> 8
        if (total >> 8) != 0:
            print("sum does not fit in 256-bit int")
        xprivate_key = i.hex()

    child_xprivate = xprivate_key
    return child_xprivate


def get_child_xpublic_key(xpublic_key, indexes: Optional[List[str]] = None,
                          path: Optional[str] = None) -> str:
    """
    Get Bytom get child xpublic key.

    :param xpublic_key: Bytom xpublic key.
    :type xpublic_key: str
    :param indexes: Bytom derivation indexes, default to ["2c000000", "99000000", "01000000", "00000000", "01000000"].
    :type indexes: list
    :param path: Bytom derivation path, default to None.
    :type path: str

    :return: str -- Bytom child xpublic key.

    >>> from pybytom.wallet.tools import get_child_xpublic_key
    >>> get_child_xpublic_key("16476b7fd68ca2acd92cfc38fa353e75d6103f828276f44d587e660a6bd7a5c5ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b")
    "91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e25803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141"
    """

    if indexes is None and path is None:
        indexes = []
    elif path is not None:
        indexes = path_to_indexes(path=path)

    for index in range(len(indexes)):
        index_bytes = get_bytes(indexes[index])
        xpublic_bytes = get_bytes(xpublic_key)
        i = bytearray(hmac.HMAC(xpublic_bytes[32:],
                                b"N" + xpublic_bytes[:32] + index_bytes,
                                digestmod=hashlib.sha512).digest())

        il, ir = i[:32], i[32:]
        bad_seed_checker(il)

        f = bytes(prune_intermediate_scalar(il))
        scalar = decodeint(f)
        f = scalarmultbase(scalar)

        p = decodepoint(xpublic_bytes[:32])
        p = edwards_add(p, f)
        public_key = encodepoint(p)

        xpublic_bytes = public_key[:32] + ir
        xpublic_key = xpublic_bytes.hex()

    child_xpublic = xpublic_key
    return child_xpublic


def get_private_key(xprivate_key: str, indexes: Optional[List[str]] = None,
                    path: Optional[str] = None) -> str:
    """
    Get Bytom private key from xprivate key. This is also the same with get_child_xprivate_key function.

    :param xprivate_key: Bytom xprivate key.
    :type xprivate_key: str
    :param indexes: Bytom derivation indexes, default to ["2c000000", "99000000", "01000000", "00000000", "01000000"].
    :type indexes: list
    :param path: Bytom derivation path, default to None.
    :type path: str

    :return: str -- Bytom private key.

    >>> from pybytom.wallet.tools import get_private_key
    >>> get_private_key("205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee51ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b")
    "e07af52746e7cccd0a7d1fba6651a6f474bada481f34b1c5bab5e2d71e36ee515803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141"
    """

    return get_child_xprivate_key(
        xprivate_key=xprivate_key, indexes=indexes, path=path)


def get_public_key(xpublic_key: str, indexes: Optional[List[str]] = None,
                   path: Optional[str] = None) -> str:
    """
    Get Bytom public key from xpublic key.

    :param xpublic_key: Bytom xpublic key.
    :type xpublic_key: str
    :param indexes: Bytom derivation indexes, default to ["2c000000", "99000000", "01000000", "00000000", "01000000"].
    :type indexes: list
    :param path: Bytom derivation path, default to None.
    :type path: str

    :return: str -- Bytom public key.

    >>> from pybytom.wallet.tools import get_public_key
    >>> get_public_key("16476b7fd68ca2acd92cfc38fa353e75d6103f828276f44d587e660a6bd7a5c5ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b")
    "91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2"
    """

    return get_child_xpublic_key(
            xpublic_key=xpublic_key, indexes=indexes, path=path)[:64]


def get_program(public_key: str) -> str:
    """
    Get Bytom control program from public key.

    :param public_key: Bytom public key.
    :type public_key: str
    :return: str -- Bytom control program.

    >>> from pybytom.wallet.tools import get_program
    >>> get_program("91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2")
    "00142cda4f99ea8112e6fa61cdd26157ed6dc408332a"
    """

    public_byte = get_bytes(public_key)
    ripemd160 = hashlib.new("ripemd160")
    ripemd160.update(public_byte)
    public_hash = ripemd160.hexdigest()
    control_program = "0014" + public_hash
    return control_program


def get_address(program: str, network: str = config["network"], vapor: bool = config["vapor"]) -> str:
    """
    Get Bytom address from program.

    :param program: Bytom control program.
    :type program: str
    :param network: Bytom network, default to mainnet.
    :type network: str
    :param vapor: Bytom sidechain vapor, defaults to False.
    :type vapor: bool

    :return: str -- Bytom address.

    >>> from pybytom.wallet.tools import get_address
    >>> get_address("00142cda4f99ea8112e6fa61cdd26157ed6dc408332a", "mainnet", False)
    "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7"
    >>> get_address("00142cda4f99ea8112e6fa61cdd26157ed6dc408332a", "mainnet", True)
    "vp1q9ndylx02syfwd7npehfxz4lddhzqsve2za23ag"
    """

    if not is_network(network=network):
        raise NetworkError(f"Invalid '{network}' network",
                           "choose only 'mainnet', 'solonet' or 'testnet' networks.")
    elif network == "mainnet" and not vapor:
        return encode("bm", 0, get_bytes(program[4:]))
    elif network == "solonet" and not vapor:
        return encode("sm", 0, get_bytes(program[4:]))
    elif network == "testnet" and not vapor:
        return encode("tm", 0, get_bytes(program[4:]))
    elif network == "mainnet" and vapor:
        return encode("vp", 0, get_bytes(program[4:]))
    elif network == "solonet" and vapor:
        return encode("sp", 0, get_bytes(program[4:]))
    elif network == "testnet" and vapor:
        return encode("tp", 0, get_bytes(program[4:]))



#!/usr/bin/env python3


def get_bytes(string):
    if isinstance(string, bytes):
        byte = string
    elif isinstance(string, str):
        byte = bytes.fromhex(string)
    else:
        raise TypeError("Agreement must be either 'bytes' or 'string'!")
    return byte


def bad_seed_checker(il, ir=False):
    if ir:
        parse_ir = str(il)
        if not parse_ir:
            raise ValueError("Bad seed, resulting in invalid key!")
    else:
        parse_il = int.from_bytes(il, "big")
        if parse_il == 0:
            raise ValueError("Bad seed, resulting in invalid key!")


def prune_root_scalar(string):
    s = bytearray(get_bytes(string=string))
    s[0] = s[0] & 248
    # clear top 3 bits
    s[31] = s[31] & 31
    # set second highest bit
    s[31] = s[31] | 64
    return s


def prune_intermediate_scalar(f):
    f = bytearray(f)
    # clear bottom 3 bits
    f[0] = f[0] & 248
    # clear 7 high bits
    f[29] = f[29] & 1
    # clear 8 bits
    f[30] = 0
    # clear 8 bits
    f[31] = 0
    return f



#!/usr/bin/env python3

from typing import (Optional, List)

from ..wallet import Wallet
from ..wallet.tools import (
    indexes_to_path, get_program, get_address
)
from ..rpc import build_transaction
from ..exceptions import (
    ClientError, NetworkError, AddressError
)
from ..utils import (
    amount_converter, is_network, is_address
)
from ..config import config
from .actions import (
    spend_wallet, control_address
)

# Bytom config
config: dict = config()


class Transaction:
    """
    Bytom Transaction class.

    :param network: Bytom network, defaults to mainnet.
    :type network: str
    :param vapor: Bytom sidechain vapor, defaults to False.
    :type vapor: bool

    :returns: Transaction -- Bytom transaction instance.
    .. note::
        Bytom has only three networks, ``mainnet``. ``solonet`` and ``testnet``.
    """

    def __init__(self, network: str = config["network"], vapor: bool = config["vapor"]):
        if not is_network(network=network):
            raise NetworkError(f"Invalid '{network}' network",
                               "choose only 'mainnet', 'solonet' or 'testnet' networks.")
        self._network: str = network
        self._vapor: bool = vapor

        self._transaction: Optional[dict] = None
        self._fee: int = config["fee"]
        self._signatures: List[List[str]] = []
        self._confirmations: int = config["confirmations"]
        self._forbid_chain_tx: bool = config["forbid_chain_tx"]

    def build_transaction(self, **kwargs):
        """
        Build Bytom transaction.

        :param kwargs: Arbitrary keyword arguments. If you do accept ``**kwargs``, make sure
        you link to documentation that describes what keywords are accepted,
        or list the keyword arguments as a definition list:
        ``address`` (str) Bytom address.
        ``inputs`` (list) Bytom transaction inputs.
        ``outputs`` (list) Bytom transaction outputs.
        ``fee`` (int) Bytom transaction fee, defaults to 10000000.
        ``confirmations`` (int) confirmations: Bytom transaction confirmations, defaults to 1.
        ``forbid_chain_tx`` (bool) forbid_chain_tx: Whether to prohibit chain transactions, defaults to False.

        :returns: Transaction -- Bytom transaction class instance.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction(address="bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", inputs=[[...], ...], outputs=[[...], ...], fee=10000000, confirmations=3)
        <pybytom.transaction.transaction.Transaction object at 0x0409DAF0>
        """

        # Checking build transaction arguments instance
        if {"address", "inputs", "outputs"} - set(kwargs.keys()):
            raise ClientError("You can't build transaction without 'address', 'inputs' and 'outputs'",
                              "default fee is 10000000 and confirmations to 1.")
        if self._vapor and not is_address(address=kwargs["address"], network=self._network, vapor=True):
            raise AddressError(f"Invalid '{kwargs['address']}' {self._network} vapor address.")
        if not self._vapor and not is_address(address=kwargs["address"], network=self._network, vapor=False):
            raise AddressError(f"Invalid '{kwargs['address']}' {self._network} address.")
        if not isinstance(kwargs["inputs"], list):
            raise TypeError("Invalid inputs instance, only takes list type.")
        if not isinstance(kwargs["outputs"], list):
            raise TypeError("Invalid outputs instance, only takes list type.")
        if "fee" in kwargs.keys() and not isinstance(kwargs["fee"], int):
            raise TypeError("Invalid fee instance, only takes integer type.")
        if "confirmations" in kwargs.keys() and not isinstance(kwargs["confirmations"], int):
            raise TypeError("Invalid confirmations instance, only takes integer type.")
        if "forbid_chain_tx" in kwargs.keys() and not isinstance(kwargs["confirmations"], int):
            raise TypeError("Invalid forbid_chain_tx instance, only takes boolean type.")

        # Setting fee, confirmations & forbid_chain_tx
        self._fee = kwargs["fee"] if "fee" in kwargs.keys() else config["fee"]
        self._confirmations = kwargs["confirmations"] \
            if "confirmations" in kwargs.keys() else config["confirmations"]
        self._forbid_chain_tx = kwargs["forbid_chain_tx"] \
            if "forbid_chain_tx" in kwargs.keys() else config["forbid_chain_tx"]

        # Building transaction
        self._transaction = build_transaction(
            address=kwargs["address"],
            transaction=dict(
                fee=str(amount_converter(self._fee, "NEU2BTM")),
                inputs=kwargs["inputs"],
                outputs=kwargs["outputs"],
                confirmations=self._confirmations,
                forbid_chain_tx=self._forbid_chain_tx
            ),
            network=self._network,
            vapor=self._vapor
        )
        return self

    def fee(self) -> int:
        """
        Bytom transaction fee.

        :returns: int -- Bytom transaction fee.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", inputs=[...], outputs=[...])
        >>> transaction.fee()
        10000000
        """

        return self._fee

    def confirmations(self) -> int:
        """
        Bytom transaction confirmations.

        :returns: int -- Bytom transaction confirmations.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", inputs=[...], outputs=[...])
        >>> transaction.confirmations()
        2
        """

        return self._confirmations

    def hash(self) -> str:
        """
        Bytom transaction hash.

        :returns: str -- Bytom transaction hash or transaction id.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", inputs=[...], outputs=[...])
        >>> transaction.hash()
        "2993414225f65390220730d0c1a356c14e91bca76db112d37366df93e364a492"
        """

        if self._transaction is None:
            raise ValueError("Transaction is none, build transaction first.")
        return self._transaction["tx"]["hash"]

    def json(self):
        """
        Bytom transaction json format.

        :returns: dict -- Bytom transaction json format.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", inputs=[...], outputs=[...])
        >>> transaction.json()
        {'hash': '2993414225f65390220730d0c1a356c14e91bca76db112d37366df93e364a492', 'status_fail': false, 'size': 379, 'submission_timestamp': 0, 'memo': "", 'inputs': [{'script': '00142cda4f99ea8112e6fa61cdd26157ed6dc408332a', 'address': 'bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7', 'asset': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 2450000000, 'type': 'spend'}], 'outputs': [{'utxo_id': '5edccebe497893c289121f9e365fdeb34c97008b9eb5a9960fe9541e7923aabc', 'script': '01642091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e220ac13c0bb1445423a641754182d53f0677cd4351a0e743e6f10b35122c3d7ea01202b9a5949f5546f63a253e41cda6bffdedb527288a7e24ed953f5c2680c70d6ff741f547a6416000000557aa888537a7cae7cac631f000000537acd9f6972ae7cac00c0', 'address': 'smart contract', 'asset': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 1000, 'type': 'control'}, {'utxo_id': 'f8cfbb692db1963be88b09c314adcc9e19d91c6c019aa556fb7cb76ba8ffa1fa', 'script': '00142cda4f99ea8112e6fa61cdd26157ed6dc408332a', 'address': 'bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7', 'asset': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 2439999000, 'type': 'control'}], 'fee': 10000000, 'balances': [{'asset': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': '-10001000'}], 'types': ['ordinary']}
        """
        if self._transaction is None:
            raise ValueError("Transaction is none, build transaction first.")
        return self._transaction["tx"]

    def raw(self) -> str:
        """
        Bytom transaction raw.

        :returns: str -- Bytom transaction raw.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", inputs=[...], outputs=[...])
        >>> transaction.raw()
        "070100010160015e7f2d7ecec3f61d30d0b2968973a3ac8448f0599ea20dce883b48c903c4d6e87fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff8091a0900901011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a22012091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e20201ad01ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffe80701880101642091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e220ac13c0bb1445423a641754182d53f0677cd4351a0e743e6f10b35122c3d7ea01202b9a5949f5546f63a253e41cda6bffdedb527288a7e24ed953f5c2680c70d6ff741f547a6416000000557aa888537a7cae7cac631f000000537acd9f6972ae7cac00c000013dffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff98dcbd8b09011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a00"
        """

        if self._transaction is None:
            raise ValueError("Transaction is none, build transaction first.")
        return self._transaction["raw_transaction"]

    def unsigned_datas(self, detail: bool = False) -> List[dict]:
        """
        Bytom unsigned transaction datas with signing instructions.

        :returns: dict -- Bytom transaction json format.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", inputs=[...], outputs=[...])
        >>> transaction.unsigned_datas()
        [{'datas': ['a7ba7ffd969e4150b36ea9f1f861d53c6f69226a3519023899d27bf1f85be3d9'], 'public_key': '91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2', 'network': 'mainnet', 'path': 'm/44/153/1/0/1'}, {'datas': ['0dbf0e408099017711a941d9a9899541afa64eb31a7118815d37e6d603b7c329'], 'public_key': '91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2', 'network': 'mainnet', 'path': 'm/44/153/1/0/1'}]
        """

        # Checking transaction
        if self._transaction is None:
            raise ValueError("Transaction is none, build transaction first.")

        unsigned_datas: List[dict] = []
        for signing_instruction in self._transaction["signing_instructions"]:
            unsigned_data = dict(datas=signing_instruction["sign_data"])
            if "pubkey" in signing_instruction and signing_instruction["pubkey"]:
                unsigned_data.setdefault("public_key", signing_instruction["pubkey"])
                if detail:
                    program = get_program(public_key=signing_instruction["pubkey"])
                    address = get_address(program=program, network=self._network)
                    unsigned_data.setdefault("program", program)
                    unsigned_data.setdefault("address", address)
                else:
                    unsigned_data.setdefault("network", self._network)
            else:
                if detail:
                    unsigned_data.setdefault("public_key", None)
                    unsigned_data.setdefault("program", None)
                    unsigned_data.setdefault("address", None)
                else:
                    unsigned_data.setdefault("network", self._network)
            if "derivation_path" in signing_instruction and signing_instruction["derivation_path"]:
                path = indexes_to_path(indexes=signing_instruction["derivation_path"])
                if detail:
                    unsigned_data.setdefault("indexes", signing_instruction["derivation_path"])
                unsigned_data.setdefault("path", path)
            else:
                if detail:
                    unsigned_data.setdefault("indexes", None)
                unsigned_data.setdefault("path", None)
            unsigned_datas.append(unsigned_data)
        return unsigned_datas

    def sign(self, private_key: Optional[str] = None, xprivate_key: Optional[str] = None,
             account: int = 1, change: bool = False, address: int = 1,
             path: str = None, indexes: List[str] = None) -> "Transaction":
        """
        Bytom sign unsigned transaction datas.

        :param private_key: Bytom private key, default to None.
        :type private_key: str
        :param xprivate_key: Bytom xprivate key, default to None.
        :type xprivate_key: str
        :param account: Bytom derivation account, defaults to 1.
        :type account: int
        :param change: Bytom derivation change, defaults to False.
        :type change: bool
        :param address: Bytom derivation address, defaults to 1.
        :type address: int
        :param path: Bytom derivation path, default to None.
        :type path: str.
        :param indexes: Bytom derivation indexes, default to None.
        :type indexes: list.

        :returns: Transaction -- Bytom transaction instance.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", inputs=[...], outputs=[...])
        >>> transaction.sign(xprivate_key)
        <pybytom.transaction.transaction.Transaction object at 0x0409DAF0>
        """

        self._signatures = []
        wallet = Wallet(network=self._network)
        if private_key and not xprivate_key:
            wallet.from_private_key(private_key=private_key)
        elif not private_key and xprivate_key:
            wallet.from_xprivate_key(xprivate_key=xprivate_key)
        for signing_instruction in self.unsigned_datas(detail=True):
            signed_data: list = []
            unsigned_datas = signing_instruction["datas"]
            if not private_key and signing_instruction["path"]:
                wallet.from_path(signing_instruction["path"])
            elif not private_key and not path and not indexes:
                wallet.from_path(f"m/44/153/{account}/{1 if change else 0}/{address}")
            elif not private_key and path:
                wallet.from_path(path)
            elif not private_key and indexes:
                wallet.from_indexes(indexes)
            for unsigned_data in unsigned_datas:
                signed_data.append(wallet.sign(unsigned_data))
            self._signatures.append(signed_data)
            wallet.clean_derivation()
        return self

    def signatures(self) -> List[List[str]]:
        """
        Bytom signed datas/signatures.

        :returns: list -- Bytom sign datas/signatures.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", inputs=[...], outputs=[...])
        >>> transaction.sign(xprivate_key="3842e3fa2af2a687e7fd67655e7a02e85bbb4ca378d4338ff590dedc7ddff447797e1e781190835138c2d1a96d0e654b625a4c019cbc64f71100be7ad1b8d4ed")
        >>> transaction.signatures()
        [[...], ...]
        """

        return self._signatures


class NormalTransaction(Transaction):
    """
    Bytom Normal transaction class.

    :param network: Bytom network, defaults to solonet.
    :type network: str
    :param vapor: Bytom sidechain vapor, defaults to False.
    :type vapor: bool

    :returns:  NormalTransaction -- Bytom normal transaction instance.
    """

    def __init__(self, network: str = config["network"], vapor: bool = config["vapor"]):
        super().__init__(network=network, vapor=vapor)

    def build_transaction(self, address: str, recipients: dict, asset: str = config["asset"],
                          fee: int = config["fee"], confirmations: int = config["confirmations"],
                          forbid_chain_tx: bool = config["forbid_chain_tx"]) -> "NormalTransaction":
        """
        Build Bytom normal transaction.

        :param address: Bytom sender blockcenter guid.
        :type address: str
        :param recipients: Recipients Bytom address and amount.
        :type recipients: dict
        :param asset: Bytom asset id, defaults to BTM asset.
        :type asset: str
        :param fee: Bytom transaction fee, defaults to 10000000.
        :type fee: int
        :param confirmations: Bytom transaction confirmations, defaults to 1.
        :type confirmations: int
        :param forbid_chain_tx: Whether to prohibit chain transactions, defaults to False.
        :type forbid_chain_tx: int

        :returns: NormalTransaction -- Bytom normal transaction instance.

        >>> from pybytom.transaction import NormalTransaction
        >>> normal_transaction = NormalTransaction(network="mainnet")
        >>> normal_transaction.build_transaction(address="bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", recipients={"bm1qtwtdhf6jmxhfhutjacmgxyv6levnkuhad67wqh": 10_000_000_000}, asset="ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff")
        <pybytom.transaction.transaction.NormalTransaction object at 0x0409DAF0>
        """

        if self._vapor and not is_address(address=address, network=self._network, vapor=True):
            raise AddressError(f"Invalid '{address}' {self._network} vapor address.")
        elif not self._vapor and not is_address(address=address, network=self._network, vapor=False):
            raise AddressError(f"Invalid '{address}' {self._network} address.")

        # Setting transaction fee and confirmations
        self._fee, self._confirmations, self._forbid_chain_tx = (
            fee, confirmations, forbid_chain_tx
        )

        inputs, outputs = [], []
        # Inputs action
        inputs.append(
            spend_wallet(
                asset=asset, amount=sum(recipients.values())
            )
        )
        # Outputs action
        for _address, _amount in recipients.items():
            outputs.append(
                control_address(
                    asset=asset, address=_address, amount=_amount,
                    vapor=self._vapor, symbol="NEU"
                )
            )

        # Building transaction
        self._transaction = build_transaction(
            address=address,
            transaction=dict(
                fee=str(amount_converter(self._fee, "NEU2BTM")),
                inputs=inputs,
                outputs=outputs,
                confirmations=self._confirmations,
                forbid_chain_tx=self._forbid_chain_tx
            ),
            network=self._network,
            vapor=self._vapor
        )
        return self


class AdvancedTransaction(Transaction):
    """
    Bytom Advanced transaction class.

    :param network: Bytom network, defaults to mainnet.
    :type network: str
    :param vapor: Bytom sidechain vapor, defaults to False.
    :type vapor: bool

    :returns: AdvancedTransaction -- Bytom advanced transaction instance.
    """

    def __init__(self, network: str = config["network"], vapor: bool = config["vapor"]):
        super().__init__(network=network, vapor=vapor)

    def build_transaction(self, address: str, inputs: list, outputs: list,
                          fee: int = config["fee"], confirmations: int = config["confirmations"],
                          forbid_chain_tx: int = config["forbid_chain_tx"]) -> "AdvancedTransaction":
        """
        Build Bytom advanced transaction.

        :param address: Bytom address.
        :type address: str
        :param inputs: Bytom transaction inputs.
        :type inputs: list
        :param outputs: Bytom transaction outputs.
        :type outputs: list
        :param fee: Bytom transaction fee, defaults to 10000000.
        :type fee: int
        :param confirmations: Bytom transaction confirmations, defaults to 1.
        :type confirmations: int
        :param forbid_chain_tx: Whether to prohibit chain transactions, defaults to False.
        :type forbid_chain_tx: int

        :returns:  AdvancedTransaction -- Bytom advanced transaction instance.

        >>> from pybytom.transaction import AdvancedTransaction
        >>> advanced_transaction = AdvancedTransaction(network="mainnet")
        >>> advanced_transaction.build_transaction("bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", [[...],...], [[...],...], 10000000, 3)
        <pybytom.transaction.transaction.AdvancedTransaction object at 0x0409DAF0>
        """

        if self._vapor and not is_address(address=address, network=self._network, vapor=True):
            raise AddressError(f"Invalid '{address}' {self._network} vapor address.")
        elif not self._vapor and not is_address(address=address, network=self._network, vapor=False):
            raise AddressError(f"Invalid '{address}' {self._network} address.")

        # Setting transaction fee and confirmations
        self._fee, self._confirmations, self._forbid_chain_tx = (
            fee, confirmations, forbid_chain_tx
        )

        # Building transaction
        self._transaction = build_transaction(
            address=address,
            transaction=dict(
                fee=str(amount_converter(self._fee, "NEU2BTM")),
                inputs=inputs,
                outputs=outputs,
                confirmations=self._confirmations,
                forbid_chain_tx=self._forbid_chain_tx
            ),
            network=self._network,
            vapor=self._vapor
        )
        return self



#!/usr/bin/env python
# coding=utf-8


class ClientError(Exception):

    def __init__(self, error_message, error_detail=None):
        self.error_message = error_message
        self.error_detail = error_detail

    def __str__(self):
        if self.error_detail:
            return f"{self.error_message}, {self.error_detail}"
        else:
            return f"{self.error_message}"


class APIError(Exception):

    def __init__(self, error_message, status_code=None):
        self.status_code = status_code
        self.error_message = error_message

    def __str__(self):
        if self.status_code:
            return f"({self.status_code}), {self.error_message}"
        else:
            return f"{self.error_message}"


class InvalidURLError(Exception):

    def __init__(self, error_message):
        self.error_message = error_message

    def __str__(self):
        return f"{self.error_message}"


class NotFoundError(Exception):

    def __init__(self, error_message, error_detail=None):
        self.error_message = error_message
        self.error_detail = error_detail

    def __str__(self):
        if self.error_detail:
            return f"{self.error_message}, {self.error_detail}"
        else:
            return f"{self.error_message}"


class AddressError(Exception):

    def __init__(self, error_message, error_detail=None):
        self.error_message = error_message
        self.error_detail = error_detail

    def __str__(self):
        if self.error_detail:
            return f"{self.error_message}, {self.error_detail}"
        return f"{self.error_message}"


class BalanceError(Exception):

    def __init__(self, error_message, error_detail=None):
        self.error_message = error_message
        self.error_detail = error_detail

    def __str__(self):
        if self.error_detail:
            return f"{self.error_message}, {self.error_detail}"
        return f"{self.error_message}"


class SymbolError(Exception):

    def __init__(self, error_message, error_detail=None):
        self.error_message = error_message
        self.error_detail = error_detail

    def __str__(self):
        if self.error_detail:
            return f"{self.error_message}, {self.error_detail}"
        return f"{self.error_message}"


class DerivationError(Exception):

    def __init__(self, error_message, error_detail=None):
        self.error_message = error_message
        self.error_detail = error_detail

    def __str__(self):
        if self.error_detail:
            return f"{self.error_message}, {self.error_detail}"
        return f"{self.error_message}"


class NetworkError(Exception):

    def __init__(self, error_message, error_detail=None):
        self.error_message = error_message
        self.error_detail = error_detail

    def __str__(self):
        if self.error_detail:
            return f"{self.error_message}, {self.error_detail}"
        return f"{self.error_message}"



#!/usr/bin/env python3

from struct import pack
from binascii import unhexlify
from ctypes import c_int64

from .opcode import (
    OP_0, OP_1, OP_DATA_1, OP_PUSHDATA1, OP_PUSHDATA2, OP_PUSHDATA4
)


class Builder:

    def __init__(self):
        self.program = bytearray()

    def add_op(self, op: int) -> "Builder":
        self.program.append(op)
        return self

    def add_bytes(self, data: bytes) -> "Builder":
        len_byte = len(data)

        if len_byte == 0:
            self.program.append(OP_0)
        elif len_byte <= 75:
            self.program.append(
                OP_DATA_1 + len_byte - 1
            )
            self.program += data
        elif len_byte < 1 << 8:
            self.program.append(OP_PUSHDATA1)
            self.program.append(len_byte)
            self.program += data
        elif len_byte < 1 << 16:
            self.program.append(OP_PUSHDATA2)
            self.program += pack("<H", len_byte)
            self.program += data
        else:
            self.program.append(OP_PUSHDATA4)
            self.program += pack("<L", len_byte)
            self.program += data
        return self

    def add_int(self, number: int) -> "Builder":
        if number == 0:
            self.program.append(OP_0)
            return self
        if 1 <= number <= 16:
            self.program.append(
                OP_1 + number - 1
            )
            return self
        # return
        return self.add_bytes(
            bytes(c_int64(number)).rstrip(b'\x00'))
    
    def add_raw_bytes(self, data: str) -> "Builder":
        self.program.append(
            unhexlify(data)
        )
        return self

    def digest(self) -> bytearray:
        return self.program

    def hex_digest(self) -> str:
        return self.program.hex()



#!/usr/bin/env python3

import hashlib
import ed25519
import hmac

from .wallet.tools import get_xpublic_key
from .libs.ed25519 import (
    sc_reduce32, decodeint, scalarmultbase, encodepoint, sc_muladd
)


def sign(private_key: str, message: str) -> str:
    """
    Sign Bytom message data by private key.

    :param private_key: Bytom private key.
    :type private_key: str.
    :param message: Message data.
    :type message: str.
    :return: str -- Bytom signed message or signature.

    >>> from pybytom.signature import sign
    >>> sign(private_key="e07af52746e7cccd0a7d1fba6651a6f474bada481f34b1c5bab5e2d71e36ee515803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141", message="1246b84985e1ab5f83f4ec2bdf271114666fd3d9e24d12981a3c861b9ed523c6")
    "f6624fea84fadccbc1bc72dc384f662468e271c4e32d846bc0a1524470549992c8ffcc3ca43891a30de4235392b0868c506ed254f0f77cc1f2b9c1a2385ddb05"
    """

    hc_hexstr = hmac.HMAC(
        b"Expand", bytes.fromhex(private_key),
        digestmod=hashlib.sha512).hexdigest()
    private = private_key[:64] + hc_hexstr[64:]
    private_bytes = bytes.fromhex(private)
    message_bytes = bytes.fromhex(message)
    data_bytes = private_bytes[32:64] + message_bytes

    message_digest = hashlib.sha512(data_bytes).digest()
    message_digest = sc_reduce32(message_digest.hex().encode())
    message_digest = bytes.fromhex(message_digest.decode())
    message_digest_reduced = message_digest[0:32]

    scalar = decodeint(message_digest_reduced)
    encoded_r = encodepoint(scalarmultbase(scalar))
    public = get_xpublic_key(private)
    public_bytes = bytes.fromhex(public)
    hram_digest_data = (
        encoded_r + public_bytes[:32] + message_bytes
    )
    hram_digest = hashlib.sha512(hram_digest_data).digest()
    hram_digest = sc_reduce32(hram_digest.hex().encode())
    hram_digest = bytes.fromhex(hram_digest.decode())
    hram_digest_reduced = hram_digest[0:32]

    sk = private_bytes[:32]
    s = sc_muladd(hram_digest_reduced.hex().encode(),
                  sk.hex().encode(), message_digest_reduced.hex().encode())
    s = bytes.fromhex(s.decode())

    signature_bytes = encoded_r + s
    signature = signature_bytes.hex()
    return signature


def verify(public_key: str, message: str, signature: str) -> bool:
    """
    Verify Bytom signature by public key.

    :param public_key: Bytom public key.
    :type public_key: str.
    :param message: Message data.
    :type message: str.
    :param signature: Signed message data.
    :type signature: str.
    :return: bool -- verified signature.

    >>> from pybytom.signature import verify
    >>> verify(public_key="91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2", message="1246b84985e1ab5f83f4ec2bdf271114666fd3d9e24d12981a3c861b9ed523c6", signature="f6624fea84fadccbc1bc72dc384f662468e271c4e32d846bc0a1524470549992c8ffcc3ca43891a30de4235392b0868c506ed254f0f77cc1f2b9c1a2385ddb05")
    True
    """

    result = False
    verifying_key = ed25519.VerifyingKey(
        public_key.encode(), encoding="hex"
    )
    try:
        verifying_key.verify(
            signature.encode(),
            bytes.fromhex(message), encoding="hex"
        )
        result = True
    except ed25519.BadSignatureError:
        result = False
    return result



import requests
import json
from _pysha3 import sha3_256
# from app.model import receiver
from pybytom.wallet.tools import get_address

# submit_transaction broadcast raw transaction
# raw_transaction_str is signed transaction,
# network_str is mainnet or testnet
# test data 1:
#   raw_transaction_str: 070100010160015e0873eddd68c4ba07c9410984799928288ae771bdccc6d974e72c95727813461fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff8094ebdc030101160014052620b86a6d5e07311d5019dffa3864ccc8a6bd630240312a052f36efb9826aa1021ec91bc6f125dd07f9c4bff87014612069527e15246518806b654d57fff8b6fe91866a19d5a2fb63a5894335fce92a7b4a7fcd340720e87ca3acdebdcad9a1d0f2caecf8ce0dbfc73d060807a210c6f225488347961402013dffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff8082eee0020116001418028ef4f8b8c278907864a1977a5ee6707b2a6b00013cffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff80b8b872011600142935e4869d0317d9701c80a02ecf888143cb9dd200
#   network_str: testnet
def submit_transaction(raw_transaction_str, network_str):
    raw_transaction_dict = {
        "transaction": raw_transaction_str
    }
    raw_transaction_json = json.dumps(raw_transaction_dict)
    headers = {
        "content-type": "application/json",
        "accept": "application/json"
    }
    if network_str == "mainnet":
        url = "https://blockmeta.com/api/v2/broadcast-transaction"
    else:
        url = "https://blockmeta.com/api/wisdom/broadcast-transaction"
    response = requests.post(url, headers=headers, data=raw_transaction_json)
    return {
        "response": response.text[:-1]
    }


def decode_raw_transaction(raw_transaction_str):
    raw_transaction_dict = {
        "raw_transaction": raw_transaction_str
    }
    raw_transaction_json = json.dumps(raw_transaction_dict)
    headers = {
        "content-type": "application/json",
        "accept": "application/json"
    }
    url = 'http://127.0.0.1:9888/decode-raw-transaction'
    response = requests.post(url, headers=headers, data=raw_transaction_json)
    return {
        "response": response.text[:-1]
    }


def get_uvarint(uvarint_str):
    uvarint_bytes = bytes.fromhex(uvarint_str)
    x, s, i = 0, 0, 0
    while True:
        b = uvarint_bytes[i]
        if b < 0x80:
            if i > 9 or i == 9 and b > 1:
                return "overflow"
            return x | int(b) << s, i + 1
        x |= int(b & 0x7f) << s
        s += 7
        i += 1


'''
get_spend_output_id create tx_input spend output id
test data 1:
  source_id_hexstr: 28b7b53d8dc90006bf97e0a4eaae2a72ec3d869873188698b694beaf20789f21
  asset_id_hexstr: ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff
  amount_int: 41250000000
  source_position_int: 0
  vmversion_int: 1
  control_program_hexstr: 00149335b1cbd4a77b78e33315a0ed10a95b12e7ca48
  spend_output_id_hexstr: f229ec6f403d586dc87aa2546bbe64c5f7b5f46eb13c6ee4823d03bc88a7cf17
'''
def get_spend_output_id(source_id_hexstr, asset_id_hexstr, amount_int, source_position_int, vmversion_int, control_program_hexstr):
    amount_hexstr = amount_int.to_bytes(8, byteorder='little').hex()
    source_position_hexstr = source_position_int.to_bytes(8, byteorder='little').hex()
    vmversion_hexstr = vmversion_int.to_bytes(8, byteorder='little').hex()
    cp_length_int = len(control_program_hexstr) // 2
    cp_length_hexstr = cp_length_int.to_bytes((cp_length_int.bit_length() + 7) // 8, byteorder='little').hex()
    sc_hexstr = source_id_hexstr + asset_id_hexstr + amount_hexstr + source_position_hexstr + vmversion_hexstr + cp_length_hexstr +  control_program_hexstr
    innerhash_bytes = sha3_256(bytes.fromhex(sc_hexstr)).digest()
    spend_bytes = b'entryid:output1:' + innerhash_bytes
    spend_output_id_hexstr = sha3_256(spend_bytes).hexdigest()
    return spend_output_id_hexstr

'''
get_input_id create tx input_id
test data 1:
    spend_output_id_hexstr: f229ec6f403d586dc87aa2546bbe64c5f7b5f46eb13c6ee4823d03bc88a7cf17
    input_id_hexstr: 6e3f378ed844b143a335e306f4ba26746157589c87e8fc8cba6463c566c56768
'''
def get_input_id(spend_output_id_hexstr):
    innerhash_bytes = sha3_256(bytes.fromhex(spend_output_id_hexstr)).digest()
    input_id_hexstr = sha3_256(b'entryid:spend1:' + innerhash_bytes).hexdigest()
    return input_id_hexstr


def get_mux_id(prepare_mux_hexstr):
    innerhash_bytes = sha3_256(bytes.fromhex(prepare_mux_hexstr)).digest()
    mux_id_hexstr = sha3_256(b'entryid:mux1:' + innerhash_bytes).hexdigest()
    return mux_id_hexstr


def get_output_id(prepare_output_id_hexstr):
    innerhash_bytes = sha3_256(bytes.fromhex(prepare_output_id_hexstr)).digest()
    output_id_hexstr = sha3_256(b'entryid:output1:' + innerhash_bytes).hexdigest()
    return output_id_hexstr


def get_tx_id(prepare_tx_id_hexstr):
    innerhash_bytes = sha3_256(bytes.fromhex(prepare_tx_id_hexstr)).digest()
    tx_id_hexstr = sha3_256(b'entryid:txheader:' + innerhash_bytes).hexdigest()
    return tx_id_hexstr


def get_issue_input_id(prepare_issue_hexstr):
    innerhash_bytes = sha3_256(bytes.fromhex(prepare_issue_hexstr)).digest()
    tx_id_hexstr = sha3_256(b'entryid:issuance1:' + innerhash_bytes).hexdigest()
    return tx_id_hexstr


def get_coinbase_input_id(prepare_coinbase_input_id_hexstr):
    innerhash_bytes = sha3_256(bytes.fromhex(prepare_coinbase_input_id_hexstr)).digest()
    coinbase_input_id_hexstr = sha3_256(b'entryid:coinbase1:' + innerhash_bytes).hexdigest()
    return coinbase_input_id_hexstr


'''
decode_raw_tx decode raw transaction
testdata 1:
    raw_transaction_str: 070100010161015f28b7b53d8dc90006bf97e0a4eaae2a72ec3d869873188698b694beaf20789f21ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff8099c4d5990100011600149335b1cbd4a77b78e33315a0ed10a95b12e7ca48630240897e2d9d24a3b5faaed0579dee7597b401491595675f897504f8945b29d836235bd2fca72a3ad0cae814628973ebcd142d9d6cc92d0b2571b69e5370a98a340c208cb7fb3086f58db9a31401b99e8c658be66134fb9034de1d5c462679270b090702013effffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff80f9f8bc98010116001406ce4b689ba026ffd3a7ca65d1d059546d4b78a000013dffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff80c6868f01011600147929ef91997c827bebf60fa608f876ea27523c4700
    network_str: solonet
    transaction: 
{
  "fee": 20000000,
  "inputs": [
    {
      "address": "sm1qjv6mrj755aah3cenzksw6y9ftvfw0jjgk0l2mw",
      "amount": 41250000000,
      "asset_definition": {},
      "asset_id": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
      "control_program": "00149335b1cbd4a77b78e33315a0ed10a95b12e7ca48",
      "input_id": "6e3f378ed844b143a335e306f4ba26746157589c87e8fc8cba6463c566c56768",
      "spent_output_id": "f229ec6f403d586dc87aa2546bbe64c5f7b5f46eb13c6ee4823d03bc88a7cf17",
      "type": "spend",
      "witness_arguments": [
        "897e2d9d24a3b5faaed0579dee7597b401491595675f897504f8945b29d836235bd2fca72a3ad0cae814628973ebcd142d9d6cc92d0b2571b69e5370a98a340c",
        "8cb7fb3086f58db9a31401b99e8c658be66134fb9034de1d5c462679270b0907"
      ]
    }
  ],
  "outputs": [
    {
      "address": "sm1qqm8yk6ym5qn0l5a8efjar5ze23k5k79qnvtslj",
      "amount": 40930000000,
      "asset_definition": {},
      "asset_id": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
      "control_program": "001406ce4b689ba026ffd3a7ca65d1d059546d4b78a0",
      "id": "74c73266730d3c6ea32e8667ef9b867068736b84be240fe9fef205fa68bb7b95",
      "position": 0,
      "type": "control"
    },
    {
      "address": "sm1q0y57lyve0jp8h6lkp7nq37rkagn4y0z8hvh6kq",
      "amount": 300000000,
      "asset_definition": {},
      "asset_id": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
      "control_program": "00147929ef91997c827bebf60fa608f876ea27523c47",
      "id": "f115a833d0c302a5006032858a7ed3987f0feb2daf2a9f849384950e4766af51",
      "position": 1,
      "type": "control"
    }
  ],
  "size": 333,
  "time_range": 0,
  "tx_id": "814a73dd57bae67c604f9cbc696cbc42035577423408cb9267136ed971e2bf63",
  "version": 1
}
'''
def decode_raw_tx(raw_transaction_str, network_str):
    tx = {
        "fee": 0,
        "inputs": [],
        "outputs": [],
        "size": 0,
        "time_range": 0,
        "tx_id": "",
        "version": 0
    }
    tx['fee'] = 0
    tx['size'] = len(raw_transaction_str) // 2
    length = 0
    offset = 2
    tx['version'], length = get_uvarint(raw_transaction_str[offset:offset+18])
    offset = offset + 2 * length
    tx['time_range'], length = get_uvarint(raw_transaction_str[offset:offset+18])
    offset = offset + 2 * length
    tx_input_amount, length = get_uvarint(raw_transaction_str[offset:offset+8])
    offset = offset + 2 * length
    prepare_mux_hexstr = (tx_input_amount).to_bytes((tx_input_amount.bit_length() + 7) // 8, 'little').hex()
    prepare_tx_id_hexstr = (tx['version']).to_bytes(8, 'little').hex() + (tx['time_range']).to_bytes(8, 'little').hex()
    for _ in range(tx_input_amount):
        _, length = get_uvarint(raw_transaction_str[offset:offset+18])
        offset = offset + 2 * length
        _, length = get_uvarint(raw_transaction_str[offset:offset+18])
        offset = offset + 2 * length
        input_type = int(raw_transaction_str[offset:offset+2], 16)
        offset += 2
        if input_type == 0: # issue
            tx_input = {
                "amount": 0,
                "asset_definition": "", # TODO:fix it!!
                "asset_id": "",
                "input_id": "",
                "issuance_program": "",
                "type": "",
                "witness_arguments": []
            }
            tx_input['type'] = "issue"
            _, length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            nonce = raw_transaction_str[offset:offset+16]
            offset += 16
            nonce_hash_hexstr = sha3_256(bytes.fromhex(nonce)).hexdigest()
            tx_input['asset_id'] = raw_transaction_str[offset:offset+64]
            offset += 64
            tx_input['amount'], length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            _, length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            asset_definition_size, length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            tx_input['asset_definition'] = bytes.fromhex(raw_transaction_str[offset:offset+2*asset_definition_size]).decode()
            offset = offset + 2 * asset_definition_size
            _, length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            issuance_program_length, length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            tx_input['issuance_program'] = raw_transaction_str[offset:offset+2*issuance_program_length]
            offset = offset + 2 * issuance_program_length
            witness_arguments_amount, length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            for _ in range(witness_arguments_amount):
                argument_length, length = get_uvarint(raw_transaction_str[offset:offset+18])
                offset = offset + 2 * length
                argument = raw_transaction_str[offset:offset+2*argument_length]
                offset = offset + 2 * argument_length
                tx_input['witness_arguments'].append(argument)
            prepare_issue_hexstr = nonce_hash_hexstr + tx_input['asset_id'] + (tx_input['amount']).to_bytes(8, byteorder='little').hex()
            tx_input['input_id'] = get_issue_input_id(prepare_issue_hexstr)
            tx['inputs'].append(tx_input)
            prepare_mux_hexstr += tx_input['input_id'] + tx_input['asset_id'] + (tx_input['amount']).to_bytes(8, byteorder='little').hex() + '0000000000000000'
            prepare_mux_hexstr += '0100000000000000' + '0151'
            mux_id_hexstr = get_mux_id(prepare_mux_hexstr)
        elif input_type == 1: # spend
            tx_input = {
                "address": "",
                "amount": 0,
                "asset_definition": {},
                "asset_id": "",
                "control_program": "",
                "input_id": "",
                "spent_output_id": "",
                "type": "",
                "witness_arguments": []
            }
            tx_input['type'] = "spend"
            _, length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            source_id = raw_transaction_str[offset:offset+64]
            offset += 64
            tx_input['asset_id'] = raw_transaction_str[offset:offset+64]
            offset += 64
            tx_input['amount'], length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            tx['fee'] += tx_input['amount']
            source_positon, length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            vmversion, length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            control_program_length, length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            tx_input['control_program'] = raw_transaction_str[offset:offset+2*control_program_length]
            offset = offset + 2 * control_program_length
            tx_input['address'] = get_address(tx_input['control_program'], network_str)
            _, length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            witness_arguments_amount, length = get_uvarint(raw_transaction_str[offset:offset+18])
            offset = offset + 2 * length
            if witness_arguments_amount == 1:
                offset = offset + 2
                tx_input['witness_arguments'] = None
            else: 
                for _ in range(witness_arguments_amount):
                    argument_length, length = get_uvarint(raw_transaction_str[offset:offset+18])
                    offset = offset + 2 * length
                    argument = raw_transaction_str[offset:offset+2*argument_length]
                    offset = offset + 2 * argument_length
                    tx_input['witness_arguments'].append(argument)
            tx_input['spent_output_id'] = get_spend_output_id(source_id, tx_input['asset_id'], tx_input['amount'], source_positon, vmversion, tx_input['control_program'])
            tx_input['input_id'] = get_input_id(tx_input['spent_output_id'])
            tx['inputs'].append(tx_input)
            prepare_mux_hexstr += tx_input['input_id'] + tx_input['asset_id'] + (tx_input['amount']).to_bytes(8, byteorder='little').hex() + '0000000000000000'
            prepare_mux_hexstr += '0100000000000000' + '0151'
            mux_id_hexstr = get_mux_id(prepare_mux_hexstr)
        elif input_type == 2: # coinbase
            tx_input = {
                "amount": 0,
                "arbitrary": "",
                "asset_definition": {},
                "asset_id": "0000000000000000000000000000000000000000000000000000000000000000",
                "input_id": "",
                "type": "",
                "witness_arguments": []
            }
            tx_input['type'] = "coinbase"
            arbitrary_length, length = get_uvarint(raw_transaction_str[offset:offset+18])
            prepare_coinbase_input_id_hexstr = raw_transaction_str[offset:offset+2*length]
            offset = offset + 2 * length
            tx_input['arbitrary'] = raw_transaction_str[offset:offset+2*arbitrary_length]
            prepare_coinbase_input_id_hexstr += tx_input['arbitrary']
            offset = offset + 2 * arbitrary_length
            tx_input['input_id'] = get_coinbase_input_id(prepare_coinbase_input_id_hexstr)
            offset = offset + 2
            tx['inputs'].append(tx_input)
            prepare_mux_hexstr += tx_input['input_id'] + 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff'
    tx_output_amount, length = get_uvarint(raw_transaction_str[offset:offset+18])
    offset = offset + 2 * length
    prepare_tx_id_hexstr += (tx_output_amount).to_bytes((tx_output_amount.bit_length() + 7) // 8, 'little').hex()
    for i in range(tx_output_amount):
        tx_output = {
            "address": "",
            "amount": 0,
            "asset_definition": {},
            "asset_id": "",
            "control_program": "",
            "id": "",
            "position": 0,
            "type": ""
        }
        tx_output['position'] = i
        _, length = get_uvarint(raw_transaction_str[offset:offset+18])
        offset = offset + 2 * length
        _, length = get_uvarint(raw_transaction_str[offset:offset+18])
        offset = offset + 2 * length
        tx_output['asset_id'] = raw_transaction_str[offset:offset+64]
        offset = offset + 64
        tx_output['amount'], length = get_uvarint(raw_transaction_str[offset:offset+18])
        if tx_input['type'] == "coinbase":
            prepare_mux_hexstr = prepare_mux_hexstr + (tx_output['amount']).to_bytes(8, byteorder='little').hex() + '0000000000000000' + '0100000000000000' + '0151'
            mux_id_hexstr = get_mux_id(prepare_mux_hexstr)
        offset = offset + 2 * length
        if tx_output['asset_id'] == 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff':
            tx['fee'] -= tx_output['amount']
        _, length = get_uvarint(raw_transaction_str[offset:offset+18])
        offset = offset + 2 * length
        control_program_length, length = get_uvarint(raw_transaction_str[offset:offset+18])
        # print(control_program_length, length)
        offset = offset + 2 * length
        tx_output['control_program'] = raw_transaction_str[offset:offset+2*control_program_length]
        offset = offset + 2 * control_program_length
        tx_output['address'] = get_address(tx_output['control_program'], network_str)
        print(tx_output['address'], offset, offset+18)
        _, length = get_uvarint(raw_transaction_str[offset:offset+18])
        offset = offset + 2 * length
        prepare_output_id_hexstr = mux_id_hexstr + tx_output['asset_id'] + (tx_output['amount']).to_bytes(8, byteorder='little').hex() + (i).to_bytes(8, byteorder='little').hex() + '0100000000000000' + (control_program_length).to_bytes((control_program_length.bit_length() + 7) // 8, 'little').hex() + tx_output['control_program']
        tx_output['id'] = get_output_id(prepare_output_id_hexstr)
        prepare_tx_id_hexstr += tx_output['id']
        tx_output['type'] = 'control'
        tx['outputs'].append(tx_output)
    if tx_input['type'] == "coinbase":
        tx['fee'] = 0
    tx['tx_id'] = get_tx_id(prepare_tx_id_hexstr)
    return tx

print(json.dumps(decode_raw_tx(
    raw_transaction_str="070100010161015f28b7b53d8dc90006bf97e0a4eaae2a72ec3d869873188698b694beaf20789f21ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff8099c4d5990100011600149335b1cbd4a77b78e33315a0ed10a95b12e7ca48630240897e2d9d24a3b5faaed0579dee7597b401491595675f897504f8945b29d836235bd2fca72a3ad0cae814628973ebcd142d9d6cc92d0b2571b69e5370a98a340c208cb7fb3086f58db9a31401b99e8c658be66134fb9034de1d5c462679270b090702013effffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff80f9f8bc98010116001406ce4b689ba026ffd3a7ca65d1d059546d4b78a000013dffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff80c6868f01011600147929ef91997c827bebf60fa608f876ea27523c4700",
    network_str="mainnet"
)))

print(json.dumps(decode_raw_tx(
    raw_transaction_str="070100010160015e76cf7c49abebba176f8ba251f70e1271ba1d876c71a2ee48799b05813ecdc07efffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff083ced00701011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a22012091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2020146ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff904e012200200ff35a6d89c16213a705a56bc4f9a736605caa981e3af5ef3a55a95cd128750f00013dffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffe088ebcb07011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a00",
    network_str="mainnet"
), indent=4))



#!/usr/bin/env python3

from binascii import unhexlify
from hashlib import (
    new, sha3_256
)

from ..libs.segwit import encode
from ..exceptions import NetworkError
from ..utils import is_network
from ..config import config
from .opcode import (
    OP_DUP, OP_HASH160, OP_SHA3, OP_EQUALVERIFY, OP_TXSIGHASH, OP_SWAP, OP_CHECKPREDICATE, OP_CHECKSIG
)
from .builder import Builder

# Bytom config
config: dict = config()


def get_public_key_hash(public_key: str) -> str:
    """
    Get Bytom public key hash.

    :param public_key: Bytom contract program(bytecode).
    :type public_key: str

    :return: str -- Public key ripemd160 hash.

    >>> from pybytom.script import get_public_key_hash
    >>> get_public_key_hash(public_key="5b5a06f6fbcb74b58ebb42293808fec6222234df6c97d7c1cff6d857a6024dc2")
    "875240ba66646d900c59dd20d843351c2fcbeedc"
    """

    ripemd160 = new("ripemd160")
    ripemd160.update(unhexlify(public_key))
    public_hash = ripemd160.hexdigest()
    return public_hash


def get_script_hash(bytecode: str) -> str:
    """
    Get Bytom Smart Contract program(bytecode) script hash.

    :param bytecode: Bytom contract program(bytecode).
    :type bytecode: str

    :return: str -- Smart Contract program script SHA3-256 hash.

    >>> from pybytom.script import get_script_hash
    >>> get_script_hash(bytecode="7baa8800c3c251547ac1")
    "e47eaf5e3a7898197068e2dd7c0209331b36a1b7a73aeb8090df816a3ce5b7d3"
    """

    return sha3_256(unhexlify(bytecode)).hexdigest()


def get_p2pkh_program(public_key_hash: str) -> str:
    """
    Get Pay to Public Key Hash (P2PKH) program.

    :param public_key_hash: Bytom public key hash.
    :type public_key_hash: str

    :return: str -- Bytom Pay to Public Key Hash (P2PKH) program.

    >>> from pybytom.script import get_p2pkh_program
    >>> get_p2pkh_program(public_key_hash="875240ba66646d900c59dd20d843351c2fcbeedc")
    "76ab14875240ba66646d900c59dd20d843351c2fcbeedc88ae7cac"
    """

    builder: Builder = Builder()
    builder.add_op(OP_DUP)
    builder.add_op(OP_HASH160)
    builder.add_bytes(unhexlify(public_key_hash))
    builder.add_op(OP_EQUALVERIFY)
    builder.add_op(OP_TXSIGHASH)
    builder.add_op(OP_SWAP)
    builder.add_op(OP_CHECKSIG)
    return builder.hex_digest()


def get_p2wpkh_program(public_key_hash: str) -> str:
    """
    Get Pay to Witness Public Key Hash (P2WPKH) program.

    :param public_key_hash: Bytom public key hash.
    :type public_key_hash: str

    :return: str -- Bytom Pay to Witness Public Key Hash (P2WPKH) program.

    >>> from pybytom.script import get_p2wpkh_program
    >>> get_p2wpkh_program(public_key_hash="875240ba66646d900c59dd20d843351c2fcbeedc")
    "0014875240ba66646d900c59dd20d843351c2fcbeedc"
    """

    builder: Builder = Builder()
    builder.add_int(0)
    builder.add_bytes(unhexlify(public_key_hash))
    return builder.hex_digest()


def get_p2wpkh_address(public_key_hash: str, network: str = config["network"], vapor: bool = config["vapor"]) -> str:
    """
    Get Pay to Witness Public Key Hash (P2WPKH) address.

    :param public_key_hash: Bytom public key hash.
    :type public_key_hash: str
    :param network: Bytom network, default to mainnet.
    :type network: str
    :param vapor: Bytom sidechain vapor, defaults to False.
    :type vapor: bool

    :return: str -- Bytom Pay to Witness Public Key Hash (P2WPKH) address.

    >>> from pybytom.script import get_p2wpkh_address
    >>> get_p2wpkh_address(public_key_hash="875240ba66646d900c59dd20d843351c2fcbeedc", vapor=False)
    "bm1qsafypwnxv3keqrzem5sdsse4rshuhmku7kpnxq"
    >>> get_p2wpkh_address(public_key_hash="875240ba66646d900c59dd20d843351c2fcbeedc", vapor=True)
    "vp1qsafypwnxv3keqrzem5sdsse4rshuhmku4h3wrk"
    """

    if not is_network(network=network):
        raise NetworkError(f"Invalid '{network}' network",
                           "choose only 'mainnet', 'solonet' or 'testnet' networks.")
    if len(unhexlify(public_key_hash)) != 20:
        raise ValueError("Invalid script hash, witness program must be 20 bytes for p2wpkh.")

    if network == "mainnet" and not vapor:
        return encode("bm", 0x00, unhexlify(public_key_hash))
    elif network == "solonet" and not vapor:
        return encode("sm", 0x00, unhexlify(public_key_hash))
    elif network == "testnet" and not vapor:
        return encode("tm", 0x00, unhexlify(public_key_hash))
    elif network == "mainnet" and vapor:
        return encode("vp", 0x00, unhexlify(public_key_hash))
    elif network == "solonet" and vapor:
        return encode("sp", 0x00, unhexlify(public_key_hash))
    elif network == "testnet" and vapor:
        return encode("tp", 0x00, unhexlify(public_key_hash))


def get_p2sh_program(script_hash: str) -> str:
    """
    Get Pay to Script Hash (P2SH) program.

    :param script_hash: Bytom contract program(bytecode) script hash.
    :type script_hash: str

    :return: str -- Bytom Pay to Script Hash (P2SH) program.

    >>> from pybytom.script import get_p2sh_program
    >>> get_p2sh_program(script_hash="e47eaf5e3a7898197068e2dd7c0209331b36a1b7a73aeb8090df816a3ce5b7d3")
    "76aa20e47eaf5e3a7898197068e2dd7c0209331b36a1b7a73aeb8090df816a3ce5b7d38808ffffffffffffffff7c00c0"
    """

    builder: Builder = Builder()
    builder.add_op(OP_DUP)
    builder.add_op(OP_SHA3)
    builder.add_bytes(unhexlify(script_hash))
    builder.add_op(OP_EQUALVERIFY)
    builder.add_int(-1)
    builder.add_op(OP_SWAP)
    builder.add_int(0)
    builder.add_op(OP_CHECKPREDICATE)
    return builder.hex_digest()


def get_p2wsh_program(script_hash: str) -> str:
    """
    Get Pay to Witness Script Hash (P2WSH) program.

    :param script_hash: Bytom contract program(bytecode) script hash.
    :type script_hash: str

    :return: str -- Bytom Pay to Witness Script Hash (P2WSH) program.

    >>> from pybytom.script import get_p2wsh_program
    >>> get_p2wsh_program(script_hash="e47eaf5e3a7898197068e2dd7c0209331b36a1b7a73aeb8090df816a3ce5b7d3")
    "0020e47eaf5e3a7898197068e2dd7c0209331b36a1b7a73aeb8090df816a3ce5b7d3"
    """

    builder: Builder = Builder()
    builder.add_int(0)
    builder.add_bytes(unhexlify(script_hash))
    return builder.hex_digest()


def get_p2wsh_address(script_hash: str, network: str = config["network"], vapor: bool = config["vapor"]) -> str:
    """
    Get Pay to Witness Script Hash (P2WSH) address.

    :param script_hash: Bytom contract program(bytecode) script hash.
    :type script_hash: str
    :param network: Bytom network, default to mainnet.
    :type network: str
    :param vapor: Bytom sidechain vapor, defaults to False.
    :type vapor: bool

    :return: str -- Bytom Pay to Witness Script Hash (P2WSH) address.

    >>> from pybytom.script import get_p2wsh_address
    >>> get_p2wsh_address(script_hash="e47eaf5e3a7898197068e2dd7c0209331b36a1b7a73aeb8090df816a3ce5b7d3", vapor=False)
    "bm1qu3l27h360zvpjurgutwhcqsfxvdndgdh5uawhqysm7qk5089klfsrrlhez"
    >>> get_p2wsh_address(script_hash="e47eaf5e3a7898197068e2dd7c0209331b36a1b7a73aeb8090df816a3ce5b7d3", vapor=True)
    "vp1qu3l27h360zvpjurgutwhcqsfxvdndgdh5uawhqysm7qk5089klfs2r64gm"
    """

    if not is_network(network=network):
        raise NetworkError(f"Invalid '{network}' network",
                           "choose only 'mainnet', 'solonet' or 'testnet' networks.")
    if len(unhexlify(script_hash)) != 32:
        raise ValueError("Invalid script hash, witness program must be 32 bytes for p2wsh.")

    if network == "mainnet" and not vapor:
        return encode("bm", 0x00, unhexlify(script_hash))
    elif network == "solonet" and not vapor:
        return encode("sm", 0x00, unhexlify(script_hash))
    elif network == "testnet" and not vapor:
        return encode("tm", 0x00, unhexlify(script_hash))
    elif network == "mainnet" and vapor:
        return encode("vp", 0x00, unhexlify(script_hash))
    elif network == "solonet" and vapor:
        return encode("sp", 0x00, unhexlify(script_hash))
    elif network == "testnet" and vapor:
        return encode("tp", 0x00, unhexlify(script_hash))



#!/usr/bin/env python3

from binascii import hexlify
from mnemonic.mnemonic import Mnemonic
from typing import (
    Optional, Union
)

import os
import unicodedata

from .libs.segwit import decode
from .exceptions import (
    SymbolError, NetworkError
)
from .config import config

# Bytom config
config: dict = config()


def generate_entropy(strength: int = 128) -> str:
    """
    Generate entropy hex string.

    :param strength: Entropy strength, default to 128.
    :type strength: int

    :returns: str -- Entropy hex string.

    >>> from pybytom.utils import generate_entropy
    >>> generate_entropy(strength=128)
    "ee535b143b0d9d1f87546f9df0d06b1a"
    """

    if strength not in [128, 160, 192, 224, 256]:
        raise ValueError(
            "Strength should be one of the following "
            "[128, 160, 192, 224, 256], but it is not (%d)."
            % strength
        )

    return hexlify(os.urandom(strength // 8)).decode()


def generate_mnemonic(language: str = "english", strength: int = 128) -> str:
    """
    Generate mnemonic words.

    :param language: Mnemonic language, default to english.
    :type language: str
    :param strength: Entropy strength, default to 128.
    :type strength: int

    :returns: str -- Mnemonic words.

    >>> from pybytom.utils import generate_mnemonic
    >>> generate_mnemonic(language="french")
    "sceptre capter séquence girafe absolu relatif fleur zoologie muscle sirop saboter parure"
    """

    if language and language not in ["english", "french", "italian", "japanese",
                                     "chinese_simplified", "chinese_traditional", "korean", "spanish"]:
        raise ValueError("Invalid language, choose only the following options 'english', 'french', 'italian', "
                         "'spanish', 'chinese_simplified', 'chinese_traditional', 'japanese or 'korean' languages.")
    if strength not in [128, 160, 192, 224, 256]:
        raise ValueError(
            "Strength should be one of the following "
            "[128, 160, 192, 224, 256], but it is not (%d)."
            % strength
        )

    return Mnemonic(language=language).generate(strength=strength)


def is_entropy(entropy: str) -> bool:
    """
    Check entropy hex string.

    :param entropy: Entropy hex string.
    :type entropy: str

    :returns: bool -- True/False.

    >>> from pybytom.utils import is_entropy
    >>> is_entropy("ee535b143b0d9d1f87546f9df0d06b1a")
    True
    """

    return len(entropy) in [32, 40, 48, 56, 64]


def is_mnemonic(mnemonic: str, language: Optional[str] = None) -> bool:
    """
    Check mnemonic words.

    :param mnemonic: Mnemonic words.
    :type mnemonic: str
    :param language: Mnemonic language, default to None.
    :type language: str

    :returns: bool -- True/False.

    >>> from pybytom.utils import is_mnemonic
    >>> is_mnemonic("sceptre capter séquence girafe absolu relatif fleur zoologie muscle sirop saboter parure")
    True
    """

    if language and language not in ["english", "french", "italian", "japanese",
                                     "chinese_simplified", "chinese_traditional", "korean", "spanish"]:
        raise ValueError("Invalid language, choose only the following options 'english', 'french', 'italian', "
                         "'spanish', 'chinese_simplified', 'chinese_traditional', 'japanese or 'korean' languages.")
    try:
        if language is None:
            for _language in ["english", "french", "italian",
                              "chinese_simplified", "chinese_traditional", "japanese", "korean", "spanish"]:
                valid = False
                if Mnemonic(language=_language).check(mnemonic=mnemonic) is True:
                    valid = True
                    break
            return valid
        else:
            return Mnemonic(language=language).check(mnemonic=mnemonic)
    except:
        return False


def get_mnemonic_language(mnemonic: str) -> Optional[str]:
    """
    Get mnemonic language.

    :param mnemonic: Mnemonic words.
    :type mnemonic: str

    :returns: str -- Mnemonic language.

    >>> from pybytom.utils import get_mnemonic_language
    >>> get_mnemonic_language("sceptre capter séquence girafe absolu relatif fleur zoologie muscle sirop saboter parure")
    "french"
    """

    if not is_mnemonic(mnemonic=mnemonic):
        raise ValueError("Invalid mnemonic words.")

    language: Optional[str] = None
    for _language in ["english", "french", "italian",
                      "chinese_simplified", "chinese_traditional", "japanese", "korean", "spanish"]:
        if Mnemonic(language=_language).check(mnemonic=mnemonic) is True:
            language = _language
            break
    return language


def get_entropy_strength(entropy: str) -> int:
    """
    Get entropy strength.

    :param entropy: Entropy hex string.
    :type entropy: str

    :returns: int -- strength.

    >>> from pybytom.utils import get_entropy_strength
    >>> get_entropy_strength("ee535b143b0d9d1f87546f9df0d06b1a")
    128
    """

    if not is_entropy(entropy=entropy):
        raise ValueError("Invalid entropy.")

    length = len(entropy)
    if length == 32:
        return 128
    elif length == 40:
        return 160
    elif length == 48:
        return 192
    elif length == 56:
        return 224
    elif length == 64:
        return 256


def get_mnemonic_strength(mnemonic: str) -> int:
    """
    Get mnemonic strength.

    :param mnemonic: Mnemonic words.
    :type mnemonic: str

    :returns: int -- strength.

    >>> from pybytom.utils import get_mnemonic_strength
    >>> get_mnemonic_strength("sceptre capter séquence girafe absolu relatif fleur zoologie muscle sirop saboter parure")
    128
    """

    if not is_mnemonic(mnemonic=mnemonic):
        raise ValueError("Invalid mnemonic words.")

    words = len(unicodedata.normalize("NFKC", mnemonic).split(" "))
    if words == 12:
        return 128
    elif words == 15:
        return 160
    elif words == 18:
        return 192
    elif words == 21:
        return 224
    elif words == 24:
        return 256


def amount_converter(amount: float, symbol: str = "NEU2BTM") -> Union[int, float]:
    """
    Bytom amount converter.

    :param amount: Bytom amount.
    :type amount: float
    :param symbol: Bytom symbol, default to NEU2BTM
    :type symbol: str

    :returns: float -- BTM asset amount.

    >>> from pybytom.utils import amount_converter
    >>> amount_converter(amount=10_000_000, symbol="NEU2BTM")
    0.1
    """

    if symbol not in ["BTM2mBTM", "BTM2NEU", "mBTM2BTM", "mBTM2NEU", "NEU2BTM", "NEU2mBTM"]:
        raise SymbolError(f"Invalid '{symbol}' symbol/type",
                          "choose only 'BTM2mBTM', 'BTM2NEU', 'mBTM2BTM', 'mBTM2NEU', 'NEU2BTM' or 'NEU2mBTM' symbols.")

    # Constant values
    BTM, mBTM, NEU = (
        config["symbols"]["BTM"],
        config["symbols"]["mBTM"],
        config["symbols"]["NEU"]
    )

    if symbol == "BTM2mBTM":
        return float((amount * mBTM) / BTM)
    elif symbol == "BTM2NEU":
        return int((amount * NEU) / BTM)
    elif symbol == "mBTM2BTM":
        return float((amount * BTM) / mBTM)
    elif symbol == "mBTM2NEU":
        return int((amount * NEU) / mBTM)
    elif symbol == "NEU2BTM":
        return float((amount * BTM) / NEU)
    elif symbol == "NEU2mBTM":
        return int((amount * mBTM) / NEU)


def is_network(network: str) -> bool:
    """
    Check network type.

    :param network: Bytom network.
    :type network: str

    :returns: bool -- Checked network.

    >>> from pybytom.utils import is_network
    >>> is_network("solonet")
    True
    """

    if isinstance(network, str):
        return network.lower() in ["mainnet", "solonet", "testnet"]
    raise TypeError("Network must be string format")


def is_address(address: str, network: Optional[str] = None, vapor: bool = config["vapor"]) -> bool:
    """
    Check Bytom address.

    :param address: Bytom address.
    :type address: str
    :param network: Bytom network, defaults to None.
    :type network: str
    :param vapor: Bytom sidechain vapor, defaults to False.
    :type vapor: bool

    :returns: bool -- Bytom valid/invalid address.

    >>> from pybytom.utils import is_address
    >>> is_address("bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", "mainnet", False)
    True
    >>> is_address("bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", "mainnet", True)
    False
    >>> is_address("vp1q9ndylx02syfwd7npehfxz4lddhzqsve2za23ag", "mainnet", True)
    True
    >>> is_address("vp1q9ndylx02syfwd7npehfxz4lddhzqsve2za23ag", "mainnet", False)
    False
    """

    if not isinstance(address, str) and not vapor:
        raise TypeError("Address must be string format")
    elif not isinstance(address, str) and vapor:
        raise TypeError("Vapor address must be string format")

    if network is None and not vapor:
        for hrp in ["bm", "sm", "tm"]:
            valid = False
            if address.startswith(hrp) and \
                    not decode(hrp, address) == (None, None):
                valid = True
                break
        return valid
    if network is None and vapor:
        for hrp in ["vp", "sp", "tp"]:
            valid = False
            if address.startswith(hrp) and \
                    not decode(hrp, address) == (None, None):
                valid = True
                break
        return valid
    elif network == "mainnet" and not vapor:
        return address.startswith("bm") and not decode("bm", address) == (None, None)
    elif network == "solonet" and not vapor:
        return address.startswith("sm") and not decode("sm", address) == (None, None)
    elif network == "testnet" and not vapor:
        return address.startswith("tm") and not decode("tm", address) == (None, None)
    elif network == "mainnet" and vapor:
        return address.startswith("vp") and not decode("vp", address) == (None, None)
    elif network == "solonet" and vapor:
        return address.startswith("sp") and not decode("sp", address) == (None, None)
    elif network == "testnet" and vapor:
        return address.startswith("tp") and not decode("tp", address) == (None, None)
    else:
        raise NetworkError(f"Invalid '{network}' network",
                           "choose only 'mainnet', 'solonet' or 'testnet' networks.")



#!/usr/bin/env python3

from binascii import unhexlify

from ..rpc import get_transaction
from ..libs.segwit import decode
from ..exceptions import NetworkError
from ..config import config

# Bytom config
config: dict = config()


def find_smart_contract_utxo(transaction_id: str, network: str = config["network"],
                             vapor: bool = config["vapor"]) -> str:
    """
    Find Bytom smart contract UTXO id by transaction id/hash.

    :param transaction_id: Bytom transaction id/hash.
    :type transaction_id: str
    :param network: Bytom network, defaults to mainnet.
    :type network: str
    :param vapor: Bytom sidechain vapor, defaults to False.
    :type vapor: bool

    :returns: str -- Smart Contract UTXO id.

    >>> from pybytom.transaction.tools import find_smart_contract_utxo
    >>> find_smart_contract_utxo("e4d4fab70a41536cf298d6f261c0a1ac169a45be47583f7240115c9059cd0d03", "mainnet", False)
    "169a45be47583f7240115c9059cd0d03e4d4fab70a41536cf298d6f261c0a1ac"
    """

    utxo_id = None
    contract_transaction = get_transaction(
        transaction_id=transaction_id, network=network, vapor=vapor
    )
    contract_outputs = contract_transaction["outputs"]
    for contract_output in contract_outputs:
        if contract_output["address"] == "smart contract":
            utxo_id = contract_output["id"]
            break
    return utxo_id


def find_p2wsh_utxo(transaction_id: str, network: str = config["network"],
                    vapor: bool = config["vapor"]) -> str:
    """
    Find Bytom segwit pay to script hash UTXO id by transaction id/hash.

    :param transaction_id: Bytom transaction id/hash.
    :type transaction_id: str
    :param network: Bytom network, defaults to mainnet.
    :type network: str
    :param vapor: Bytom sidechain vapor, defaults to False.
    :type vapor: bool

    :returns: str -- Pay to Witness Secript Hash (P2WSH) UTXO id.

    >>> from pybytom.transaction.tools import find_p2wsh_utxo
    >>> find_p2wsh_utxo("0a1ac169a45be47583f72401e4d4fab70a41536cf298d6f261c15c9059cd0d03", "mainnet", False)
    "cd0d03e4d4fab70a41536cf298d6f261c0a1ac169a45be47583f7240115c9059"
    """

    utxo_id = None
    if network == "mainnet":
        hrp = "vp" if vapor else "bm"
    elif network == "solonet":
        hrp = "sp" if vapor else "sm"
    elif network == "testnet":
        hrp = "tp" if vapor else "tm"
    else:
        raise NetworkError(f"Invalid '{network}' network",
                           "choose only 'mainnet', 'solonet' or 'testnet' networks.")
    contract_transaction = get_transaction(
        transaction_id=transaction_id, network=network, vapor=vapor
    )
    contract_outputs = contract_transaction["outputs"]
    for contract_output in contract_outputs:
        _, address_hash = decode(hrp, contract_output["address"])
        if address_hash is not None and \
                len(unhexlify(bytearray(address_hash).hex())) == 32:  # deep
            utxo_id = contract_output["id"]
            break
    return utxo_id



#!/usr/bin/env python3

from ..utils import (
    amount_converter, is_address
)
from ..exceptions import (
    SymbolError, AddressError
)
from ..config import config

# Bytom config
config: dict = config()


def spend_utxo(utxo: str) -> dict:
    """
    Get spend UTXO action.

    :param utxo: Bytom utxo id.
    :type utxo: str

    :returns: dict -- Bytom spend utxo action.

    >>> from pybytom.transaction.actions import spend_utxo
    >>> spend_utxo("169a45be47583f7240115c9059cd0d03e4d4fab70a41536cf298d6f261c0a1ac")
    {'type': 'spend_utxo, 'output': '169a45be47583f7240115c9059cd0d03e4d4fab70a41536cf298d6f261c0a1ac'}
    """

    return dict(type=str("spend_utxo"), output=utxo)


def spend_wallet(amount: float, asset: str, symbol: str = "NEU") -> dict:
    """
    Get spend wallet action.

    :param amount: Bytom amount.
    :type amount: float
    :param asset: Bytom asset.
    :type asset: str
    :param symbol: Bytom symbol, default to NEU
    :type symbol: str

    :returns: dict -- Bytom spend wallet action.

    >>> from pybytom.transaction.actions import spend_wallet
    >>> spend_wallet(10_000_000, "41536cf298d6f261c0a1ac169a45be47583f7240115c9059cd0d03e4d4fab70a")
    {'type': 'spend_wallet', 'amount': '0.1', 'asset': '41536cf298d6f261c0a1ac169a45be47583f7240115c9059cd0d03e4d4fab70a'}
    """

    if symbol.startswith("BTM"):
        amount = amount
    elif symbol.startswith("mBTM"):
        amount = amount_converter(amount, symbol="mBTM2BTM")
    elif symbol.startswith("NEU"):
        amount = amount_converter(amount, symbol="NEU2BTM")
    else:
        raise SymbolError(f"Invalid '{symbol}' symbol/type",
                          "choose only 'BTM', 'mBTM' or 'NEU' symbols.")

    return dict(
        type=str("spend_wallet"),
        amount=str(amount), asset=asset
    )


def control_program(amount: float, asset: str, program: str, symbol: str = "NEU") -> dict:
    """
    Get control program action.

    :param amount: Bytom amount.
    :type amount: float
    :param asset: Bytom asset.
    :type asset: str
    :param program: Bytom control program.
    :type program: str
    :param symbol: Bytom symbol, default to NEU
    :type symbol: str

    :returns: dict -- Bytom control program action.

    >>> from pybytom.transaction.actions import control_program
    >>> control_program(10_000_000, "41536cf298d6f261c0a1ac169a45be47583f7240115c9059cd0d03e4d4fab70a", "0020e47eaf5e3a7898197068e2dd7c0209331b36a1b7a73aeb8090df816a3ce5b7d3")
    {'type': 'control_program', 'amount': '0.1', 'asset': '41536cf298d6f261c0a1ac169a45be47583f7240115c9059cd0d03e4d4fab70a', 'control_program': '0020e47eaf5e3a7898197068e2dd7c0209331b36a1b7a73aeb8090df816a3ce5b7d3'}
    """

    if symbol.startswith("BTM"):
        amount = amount
    elif symbol.startswith("mBTM"):
        amount = amount_converter(amount, symbol="mBTM2BTM")
    elif symbol.startswith("NEU"):
        amount = amount_converter(amount, symbol="NEU2BTM")
    else:
        raise SymbolError(f"Invalid '{symbol}' symbol/type",
                          "choose only 'BTM', 'mBTM' or 'NEU' symbols.")

    return dict(
        type=str("control_program"),
        amount=str(amount), asset=asset,
        control_program=program
    )


def control_address(amount: float, asset: str, address: str,
                    vapor: bool = config["vapor"], symbol: str = "NEU") -> dict:
    """
    Get control address action.

    :param amount: Bytom amount.
    :type amount: float
    :param asset: Bytom asset.
    :type asset: str
    :param address: Bytom address.
    :type address: str
    :param symbol: Bytom symbol, default to NEU
    :type symbol: str
    :param vapor: Bytom sidechain vapor, defaults to False.
    :type vapor: bool

    :returns: dict -- Bytom control address action.

    >>> from pybytom.transaction.actions import control_address
    >>> control_address(10_000_000, "41536cf298d6f261c0a1ac169a45be47583f7240115c9059cd0d03e4d4fab70a", "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7")
    {'type': 'control_address', 'amount': '0.1', 'asset': '41536cf298d6f261c0a1ac169a45be47583f7240115c9059cd0d03e4d4fab70a', 'address': 'bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7'}
    """

    if vapor and not is_address(address=address, vapor=True):
        raise AddressError(f"Invalid recipient '{address}' vapor address.")
    elif not vapor and not is_address(address=address, vapor=False):
        raise AddressError(f"Invalid recipient '{address}' address.")

    if symbol.startswith("BTM"):
        amount = amount
    elif symbol.startswith("mBTM"):
        amount = amount_converter(amount, symbol="mBTM2BTM")
    elif symbol.startswith("NEU"):
        amount = amount_converter(amount, symbol="NEU2BTM")
    else:
        raise SymbolError(f"Invalid '{symbol}' symbol/type",
                          "choose only 'BTM', 'mBTM' or 'NEU' symbols.")

    return dict(
        type=str("control_address"),
        amount=str(amount), asset=asset, address=address
    )



#!/usr/bin/env python3

from typing import List

from .script import (
    get_public_key_hash, get_script_hash,
    get_p2pkh_program, get_p2sh_program,
    get_p2wpkh_program, get_p2wsh_program,
    get_p2wpkh_address, get_p2wsh_address
)

__all__: List[str] = [
    "get_public_key_hash", "get_script_hash",
    "get_p2wpkh_program", "get_p2wsh_program",
    "get_p2pkh_program", "get_p2sh_program",
    "get_p2wpkh_address", "get_p2wsh_address"
]
